# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing flow solutions in laplace space.

.. currentmodule:: anaflow.flow.laplace

The following functions are provided

.. autosummary::
   grf_laplace
"""
# pylint: disable=C0103
import warnings

import numpy as np
from scipy.special import kv, iv, gamma, erfcx
from pentapy import solve
from anaflow.tools.special import sph_surf

__all__ = ["grf_laplace"]


def constant(s):
    """Constant pumping."""
    return 1.0 / s


def periodic(s, a=0):
    """
    Periodic pumping.

    Q(t) = Q * cos(a * t)
    """
    if np.isclose(a, 0):
        return constant(s)
    return 1.0 / (s + a ** 2 / s)


def slug(s):
    """Slug test."""
    return np.ones_like(s)


def interval(s, a=np.inf):
    """Interval pumping in [0, t]."""
    if np.isposinf(a):
        return constant(s)
    return (1.0 - np.exp(-s * a)) / s


def accruing(s, a=0):
    """Accruing pumping with time scale t."""
    return erfcx((s * a) / 2.0) / s


PUMP_COND = {0: constant, 1: periodic, 2: slug, 3: interval, 4: accruing}


def grf_laplace(
    s,
    rad=None,
    S_part=None,
    K_part=None,
    R_part=None,
    dim=2,
    lat_ext=1.0,
    rate=None,
    K_well=None,
    cut_off_prec=1e-20,
    cond=0,
    cond_kw=None,
):
    """
    The extended GRF-model for transient flow in Laplace-space.

    The General Radial Flow (GRF) Model allowes fractured dimensions for
    transient flow under a pumping condition in a confined aquifer.
    The solutions assumes concentric annuli around the pumpingwell,
    where each annulus has its own conductivity and storativity value.

    Parameters
    ----------
    s : :class:`numpy.ndarray`
        Array with all Laplace-space-points
        where the function should be evaluated
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    S_part : :class:`numpy.ndarray` of length N
        Given storativity values for each disk
    K_part : :class:`numpy.ndarray` of length N
        Given conductivity values for each disk
    R_part : :class:`numpy.ndarray` of length N+1
        Given radii separating the disks as well as starting- and endpoints
    dim : :class:`float`
        Flow dimension. Default: 3
    lat_ext : :class:`float`
        The lateral extend of the flow-domain, used in `L^(3-dim)`. Default: 1
    rate : :class:`float`
        Pumpingrate at the well
    K_well : :class:`float`, optional
        Conductivity at the well. Default: ``K_part[0]``
    cut_off_prec : :class:`float`, optional
        Define a cut-off precision for the calculation to select the disks
        included in the calculation. Default ``1e-20``
    cond : :class:`int`, optional
        Type of the pumping condition:

            * 0 : constant
            * 1 : periodic (needs "w" as cond_kw)
            * 2 : slug (rate will be interpreted as slug-volume)
            * 3 : interval (needs "t" as cond_kw)
            * callable: laplace-transformation of the transient pumping-rate

        Default: 0
    cond_kw : :class:`dict` optional
        Keyword args for the pumping condition. Default: None

    Returns
    -------
    grf_laplace : :class:`numpy.ndarray`
        Array with all values in laplace-space

    Examples
    --------
    >>> grf_laplace([5,10],[1,2,3],[1e-3,1e-3],[1e-3,2e-3],[0,2,10], 2, 1, -1)
    array([[-2.71359196e+00, -1.66671965e-01, -2.82986917e-02],
           [-4.58447458e-01, -1.12056319e-02, -9.85673855e-04]])
    """
    cond_kw = {} if cond_kw is None else cond_kw
    cond = cond if callable(cond) else PUMP_COND[cond]
    # ensure that input is treated as arrays
    s = np.squeeze(s).reshape(-1)
    rad = np.squeeze(rad).reshape(-1)
    S_part = np.squeeze(S_part).reshape(-1)
    K_part = np.squeeze(K_part).reshape(-1)
    R_part = np.squeeze(R_part).reshape(-1)
    # the dimension is used by nu in the literature (See Barker 88)
    dim = float(dim)
    nu = 1.0 - dim / 2.0
    nu1 = nu - 1
    # the lateral extend is a bit subtle in fractured dimension
    lat_ext = float(lat_ext)
    rate = float(rate)
    # get the number of partitions
    parts = len(K_part)
    # set the conductivity at the well
    K_well = K_part[0] if K_well is None else float(K_well)
    # check the input
    if not len(R_part) - 1 == len(S_part) == len(K_part) > 0:
        raise ValueError("R_part, S_part and K_part need matching lengths.")
    if R_part[0] < 0.0:
        raise ValueError("The wellradius needs to be >= 0.")
    if not all([r1 < r2 for r1, r2 in zip(R_part[:-1], R_part[1:])]):
        raise ValueError("The radii values need to be sorted.")
    if not np.min(rad) > R_part[0] or np.max(rad) > R_part[-1]:
        raise ValueError("The given radii need to be in the given range.")
    if not all([con > 0 for con in K_part]):
        raise ValueError("The Conductivity needs to be positiv.")
    if not all([stor > 0 for stor in S_part]):
        raise ValueError("The Storage needs to be positiv.")
    if not dim > 0.0 or dim > 3.0:
        raise ValueError("The dimension needs to be positiv and <= 3.")
    if not lat_ext > 0.0:
        raise ValueError("The lateral extend needs to be positiv.")
    if not K_well > 0:
        raise ValueError("The well conductivity needs to be positiv.")

    # initialize the result
    res = np.zeros(s.shape + rad.shape)
    # the first sqrt of the diffusivity values
    diff_sr0 = np.sqrt(S_part[0] / K_part[0])
    # set the general pumping-condtion depending on the well-radius
    if R_part[0] > 0.0:
        Qs = -(s ** (-0.5)) / diff_sr0 * R_part[0] ** nu1 * cond(s, **cond_kw)
    else:
        Qs = -((2 / diff_sr0) ** nu) * s ** (-nu / 2) * cond(s, **cond_kw)

    # if there is a homgeneouse aquifer, compute the result by hand
    if parts == 1:
        # initialize the equation system
        V = np.zeros(2, dtype=float)
        M = np.array([[-gamma(1 - nu), 2.0 / gamma(nu)], [0, 1]])

        for si, se in enumerate(s):
            Cs = np.sqrt(se) * diff_sr0
            # set the pumping-condition at the well
            V[0] = Qs[si]
            # incorporate the boundary-conditions
            if R_part[0] > 0.0:
                M[0, :] = [-kv(nu1, Cs * R_part[0]), iv(nu1, Cs * R_part[0])]
            if R_part[-1] < np.inf:
                M[1, :] = [kv(nu, Cs * R_part[-1]), iv(nu, Cs * R_part[-1])]
            else:
                M[0, 1] = 0  # Bs is 0 in this case either way
            # solve the equation system
            As, Bs = np.linalg.solve(M, V)

            # calculate the head
            for ri, re in enumerate(rad):
                if re < R_part[-1]:
                    res[si, ri] = re ** nu * (
                        As * kv(nu, Cs * re) + Bs * iv(nu, Cs * re)
                    )

    # if there is more than one partition, create an equation system
    else:
        # initialize LHS and RHS for the linear equation system
        # Mb is the banded matrix for the Eq-System
        V = np.zeros(2 * parts)
        Mb = np.zeros((5, 2 * parts))
        X = np.zeros(2 * parts)
        # set the standard boundary conditions for rwell=0.0 and rinf=np.inf
        Mb[2, 0] = -gamma(1 - nu)
        Mb[1, 1] = 2.0 / gamma(nu)
        Mb[2, -1] = 1.0

        # calculate the consecutive fractions of the conductivities
        Kfrac = K_part[:-1] / K_part[1:]
        # calculate the square-root of the diffusivities
        difsr = np.sqrt(S_part / K_part)
        # calculate a temporal substitution (factor from mass-conservation)
        tmp = Kfrac * difsr[:-1] / difsr[1:]
        # match the radii to the different disks
        pos = np.searchsorted(R_part, rad) - 1

        # iterate over the laplace-variable
        for si, se in enumerate(s):
            Cs = np.sqrt(se) * difsr

            # set the pumping-condition at the well
            # --> implement other pumping conditions
            V[0] = Qs[si]

            # generate the equation system as banded matrix
            for i in range(parts - 1):
                Mb[0, 2 * i + 3] = -iv(nu, Cs[i + 1] * R_part[i + 1])
                Mb[1, 2 * i + 2 : 2 * i + 4] = [
                    -kv(nu, Cs[i + 1] * R_part[i + 1]),
                    -iv(nu1, Cs[i + 1] * R_part[i + 1]),
                ]
                Mb[2, 2 * i + 1 : 2 * i + 3] = [
                    iv(nu, Cs[i] * R_part[i + 1]),
                    kv(nu1, Cs[i + 1] * R_part[i + 1]),
                ]
                Mb[3, 2 * i : 2 * i + 2] = [
                    kv(nu, Cs[i] * R_part[i + 1]),
                    tmp[i] * iv(nu1, Cs[i] * R_part[i + 1]),
                ]
                Mb[4, 2 * i] = -tmp[i] * kv(nu1, Cs[i] * R_part[i + 1])

            # set the boundary-conditions if needed
            if R_part[0] > 0.0:
                Mb[2, 0] = -kv(nu1, Cs[0] * R_part[0])
                Mb[1, 1] = iv(nu1, Cs[0] * R_part[0])
            if R_part[-1] < np.inf:
                Mb[-2, -2] = kv(nu, Cs[-1] * R_part[-1])
                Mb[2, -1] = iv(nu, Cs[-1] * R_part[-1])
            else:  # erase the last row, since X[-1] will be 0
                Mb[0, -1] = 0
                Mb[1, -1] = 0

            # find first disk which has no impact
            Mb_cond = np.max(np.abs(Mb), axis=0)
            Mb_cond_lo = Mb_cond < cut_off_prec
            Mb_cond_hi = Mb_cond > 1 / cut_off_prec
            Mb_cond = np.logical_or(Mb_cond_lo, Mb_cond_hi)
            cond = np.where(Mb_cond)[0]
            found = cond.shape[0] > 0
            first = cond[0] // 2 if found else parts

            # initialize coefficients
            X[2 * first :] = 0.0
            # only the first disk has an impact
            if first <= 1:
                M_sgl = np.eye(2, dtype=float)
                M_sgl[:, 0] = Mb[2:4, 0]
                M_sgl[:, 1] = Mb[1:3, 1]
                # solve the equation system
                try:
                    X[:2] = np.linalg.solve(M_sgl, V[:2])
                except np.linalg.LinAlgError:
                    # set 0 if matrix singular
                    X[:2] = 0
            elif first > 1:
                # shrink the matrix
                M_sgl = Mb[:, : 2 * first]
                if first < parts:
                    M_sgl[-1, -1] = 0
                    M_sgl[-2, -1] = 0
                    M_sgl[-1, -2] = 0
                X[: 2 * first] = solve(
                    M_sgl, V[: 2 * first], is_flat=True, index_row_wise=False
                )
            np.nan_to_num(X, copy=False)

            # calculate the head (ignore small values)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)
                k0_sub = X[2 * pos] * kv(nu, Cs[pos] * rad)
                k0_sub[np.abs(X[2 * pos]) < cut_off_prec] = 0
                i0_sub = X[2 * pos + 1] * iv(nu, Cs[pos] * rad)
                i0_sub[np.abs(X[2 * pos + 1]) < cut_off_prec] = 0
                res[si, :] = rad ** nu * (k0_sub + i0_sub)

    # set problematic values to 0
    # --> the algorithm tends to violate small values,
    #     therefore this approach is suitable
    np.nan_to_num(res, copy=False)
    # scale to pumpingrate
    res *= rate / (K_well * sph_surf(dim) * lat_ext ** (3.0 - dim))

    return res


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing flow solutions in homogeneous aquifers.

.. currentmodule:: anaflow.flow.homogeneous

The following functions are provided

.. autosummary::
   thiem
   theis
   grf
"""
# pylint: disable=C0103
import numpy as np

from anaflow.tools.special import well_solution, grf_solution
from anaflow.flow.ext_grf_model import ext_grf, ext_grf_steady

__all__ = ["thiem", "theis", "grf"]


###############################################################################
# Thiem-solution
###############################################################################


def thiem(rad, r_ref, transmissivity, rate=-1e-4, h_ref=0.0):
    """
    The Thiem solution.

    The Thiem solution for steady-state flow under a pumping condition
    in a confined and homogeneous aquifer.
    This solution was presented in [Thiem1906]_.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated.
    r_ref : :class:`float`
        Reference radius with known head (see `h_ref`).
    transmissivity : :class:`float`
        Transmissivity of the aquifer.
    rate : :class:`float`, optional
        Pumpingrate at the well. Default: -1e-4
    h_ref : :class:`float`, optional
        Reference head at the reference-radius `r_ref`. Default: ``0.0``

    Returns
    -------
    head : :class:`numpy.ndarray`
        Array with all heads at the given radii.

    References
    ----------
    .. [Thiem1906] Thiem, G.,
       ''Hydrologische Methoden, J.M. Gebhardt'', Leipzig, 1906.

    Notes
    -----
    The parameters ``rad``, ``r_ref`` and ``transmissivity``
    will be checked for positivity.
    If you want to use cartesian coordiantes, just use the formula
    ``r = sqrt(x**2 + y**2)``

    Examples
    --------
    >>> thiem([1,2,3], 10, 0.001, -0.001)
    array([-0.3664678 , -0.25615   , -0.19161822])
    """
    return ext_grf_steady(rad, r_ref, transmissivity, 2, 1, rate, h_ref)


###############################################################################
# Theis-solution
###############################################################################


def theis(
    time,
    rad,
    storage,
    transmissivity,
    rate=-1e-4,
    r_well=0.0,
    r_bound=np.inf,
    h_bound=0.0,
    struc_grid=True,
    lap_kwargs=None,
):
    """
    The Theis solution.

    The Theis solution for transient flow under a pumping condition
    in a confined and homogeneous aquifer.
    This solution was presented in [Theis35]_.

    Parameters
    ----------
    time : :class:`numpy.ndarray`
        Array with all time-points where the function should be evaluated
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    storage : :class:`float`
        Storage coefficient of the aquifer.
    conductivity : :class:`float`
        Conductivity of the aquifer.
    rate : :class:`float`, optional
        Pumpingrate at the well. Default: -1e-4
    r_well : :class:`float`, optional
        Inner radius of the pumping-well. Default: ``0.0``
    r_bound : :class:`float`, optional
        Radius of the outer boundariy of the aquifer. Default: ``np.inf``
    h_bound : :class:`float`, optional
        Reference head at the outer boundary, as well as initial condition.
        Default: ``0.0``
    struc_grid : :class:`bool`, optional
        If this is set to ``False``, the `rad` and `time` array will be merged
        and interpreted as single, r-t points. In this case they need to have
        the same shapes. Otherwise a structured r-t grid is created.
        Default: ``True``
    lap_kwargs : :class:`dict` or :any:`None` optional
        Dictionary for :any:`get_lap_inv` containing `method` and
        `method_dict`. The default is equivalent to
        ``lap_kwargs = {"method": "stehfest", "method_dict": None}``.
        Default: :any:`None`

    Returns
    -------
    head : :class:`numpy.ndarray`
        Array with all heads at the given radii and time-points.

    References
    ----------
    .. [Theis35] Theis, C.,
       ''The relation between the lowering of the piezometric surface and the
       rate and duration of discharge of a well using groundwater storage'',
       Trans. Am. Geophys. Union, 16, 519-524, 1935
    """
    if np.isclose(r_well, 0) and np.isposinf(r_bound) and lap_kwargs is None:
        return well_solution(time, rad, storage, transmissivity, rate)
    return ext_grf(
        time=time,
        rad=rad,
        S_part=[storage],
        K_part=[transmissivity],
        R_part=[r_well, r_bound],
        dim=2,
        lat_ext=1,
        rate=rate,
        K_well=None,
        h_bound=h_bound,
        lap_kwargs=lap_kwargs,
        struc_grid=struc_grid,
    )


def grf(
    time,
    rad,
    storage,
    conductivity,
    dim=2,
    lat_ext=1.0,
    rate=-1e-4,
    r_well=0.0,
    r_bound=np.inf,
    h_bound=0.0,
    struc_grid=True,
    lap_kwargs=None,
):
    """
    The general radial flow (GRF) model for a pumping test.

    This solution was  presented in [Barker88]_.

    Parameters
    ----------
    time : :class:`numpy.ndarray`
        Array with all time-points where the function should be evaluated.
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated.
    storage : :class:`float`
        Storage coefficient of the aquifer.
    conductivity : :class:`float`
        Conductivity of the aquifer.
    dim : :class:`float`, optional
        Fractional dimension of the aquifer. Default: ``2.0``
    lat_ext : :class:`float`, optional
        Lateral extend of the aquifer. Default: ``1.0``
    rate : :class:`float`, optional
        Pumpingrate at the well. Default: -1e-4
    r_well : :class:`float`, optional
        Inner radius of the pumping-well. Default: ``0.0``
    r_bound : :class:`float`, optional
        Radius of the outer boundary of the aquifer. Default: ``np.inf``
    h_bound : :class:`float`, optional
        Reference head at the outer boundary, as well as initial condition.
        Default: ``0.0``
    struc_grid : :class:`bool`, optional
        If this is set to "False", the "rad" and "time" array will be merged
        and interpreted as single, r-t points. In this case they need to have
        the same shapes. Otherwise a structured r-t grid is created.
        Default: ``True``
    lap_kwargs : :class:`dict` or :any:`None` optional
        Dictionary for :any:`get_lap_inv` containing `method` and
        `method_dict`. The default is equivalent to
        ``lap_kwargs = {"method": "stehfest", "method_dict": None}``.
        Default: :any:`None`

    Returns
    -------
    head : :class:`numpy.ndarray`
        Array with all heads at the given radii and time-points.

    References
    ----------
    .. [Barker88] Barker, J.
       ''A generalized radial flow model for hydraulic tests
       in fractured rock.'',
       Water Resources Research 24.10, 1796-1804, 1988
    """
    if np.isclose(r_well, 0) and np.isposinf(r_bound) and lap_kwargs is None:
        return grf_solution(
            time=time,
            rad=rad,
            storage=storage,
            conductivity=conductivity,
            dim=dim,
            lat_ext=lat_ext,
            rate=rate,
            h_bound=h_bound,
            struc_grid=struc_grid,
        )
    return ext_grf(
        time=time,
        rad=rad,
        S_part=[storage],
        K_part=[conductivity],
        R_part=[r_well, r_bound],
        dim=dim,
        lat_ext=lat_ext,
        rate=rate,
        K_well=None,
        h_bound=h_bound,
        struc_grid=struc_grid,
        lap_kwargs=lap_kwargs,
    )


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as plt
from anaflow import theis


time = [10, 100, 1000]
rad = np.geomspace(0.1, 10)

head = theis(time=time, rad=rad, storage=1e-4, transmissivity=1e-4, rate=-1e-4)

for i, step in enumerate(time):
    plt.plot(rad, head[i], label="Theis(t={})".format(step))

plt.legend()
plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as plt
import matplotlib.gridspec as gridspec
from anaflow import ext_grf, ext_grf_steady
from anaflow.tools import specialrange_cut, annular_hmean, step_f


def cond(rad, K_far, K_well, len_scale):
    """Conductivity with linear increase from K_well to K_far."""
    return np.minimum(np.abs(rad) / len_scale, 1.0) * (K_far - K_well) + K_well


time_labels = ["10 s", "100 s", "1000 s"]
time = [10, 100, 1000]
rad = np.geomspace(0.1, 6)
S = 1e-4
K_well = 1e-5
K_far = 1e-4
len_scale = 5.0
rate = -1e-4
dim = 1.5

cut_off = len_scale
parts = 30
r_well = 0.0
r_bound = 50.0

# calculate a disk-distribution of "trans" by calculating harmonic means
R_part = specialrange_cut(r_well, r_bound, parts, cut_off)
K_part = annular_hmean(cond, R_part, ann_dim=dim, K_far=K_far, K_well=K_well, len_scale=len_scale)
S_part = np.full_like(K_part, S)
# calculate transient and steady heads
head1 = ext_grf(time, rad, S_part, K_part, R_part, dim=dim, rate=rate)
head2 = ext_grf_steady(rad, r_bound, cond, dim=dim, rate=-1e-4, K_far=K_far, K_well=K_well, len_scale=len_scale)

# plotting
gs = gridspec.GridSpec(2, 1, height_ratios=[1, 3])
ax1 = plt.subplot(gs[0])
ax2 = plt.subplot(gs[1], sharex=ax1)
time_ticks=[]
for i, step in enumerate(time):
    label = "Transient" if i == 0 else None
    ax2.plot(rad, head1[i], label=label, color="C"+str(i))
    time_ticks.append(head1[i][-1])

ax2.plot(rad, head2, label="Steady", color="k", linestyle=":")

rad_lin = np.linspace(rad[0], rad[-1], 1000)
ax1.plot(rad_lin, step_f(rad_lin, R_part, K_part), label="step Conductivity")
ax1.plot(rad_lin, cond(rad_lin, K_far, K_well, len_scale), label="Conductivity")
ax1.set_yticks([K_well, K_far])
ax1.set_ylabel(r"$K$ in $[\frac{m}{s}]$")
plt.setp(ax1.get_xticklabels(), visible=False)
ax1.legend()
ax2.set_xlabel("r in [m]")
ax2.set_ylabel("h in [m]")
ax2.legend()
ax2.set_xlim([0, rad[-1]])
ax3 = ax2.twinx()
ax3.set_yticks(time_ticks)
ax3.set_yticklabels(time_labels)
ax3.set_ylim(ax2.get_ylim())

plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing flow-solutions for the groundwater flow equation.

Subpackages
^^^^^^^^^^^

.. currentmodule:: anaflow.flow

.. autosummary::
    homogeneous
    heterogeneous
    ext_grf_model
    laplace

Solutions
^^^^^^^^^

Homogeneous
~~~~~~~~~~~

.. currentmodule:: anaflow.flow.homogeneous

Solutions for homogeneous aquifers

.. autosummary::
   thiem
   theis
   grf

Heterogeneous
~~~~~~~~~~~~~

.. currentmodule:: anaflow.flow.heterogeneous

Solutions for heterogeneous aquifers

.. autosummary::
   ext_thiem_2d
   ext_thiem_3d
   ext_thiem_tpl
   ext_thiem_tpl_3d
   ext_theis_2d
   ext_theis_3d
   ext_theis_tpl
   ext_thiem_tpl_3d
   neuman2004
   neuman2004_steady

Extended GRF
~~~~~~~~~~~~

.. currentmodule:: anaflow.flow.ext_grf_model

The extended general radial flow model.

.. autosummary::
   ext_grf
   ext_grf_steady
"""
from anaflow.flow.homogeneous import thiem, theis, grf
from anaflow.flow.heterogeneous import (
    ext_thiem_2d,
    ext_thiem_3d,
    ext_thiem_tpl,
    ext_thiem_tpl_3d,
    ext_theis_2d,
    ext_theis_3d,
    ext_theis_tpl,
    ext_theis_tpl_3d,
    neuman2004,
    neuman2004_steady,
)
from anaflow.flow.ext_grf_model import ext_grf, ext_grf_steady

__all__ = [
    "thiem",
    "theis",
    "grf_model",
    "ext_thiem_2d",
    "ext_thiem_3d",
    "ext_thiem_tpl",
    "ext_thiem_tpl_3d",
    "ext_theis_2d",
    "ext_theis_3d",
    "ext_theis_tpl",
    "ext_theis_tpl_3d",
    "neuman2004",
    "neuman2004_steady",
    "grf",
    "ext_grf",
    "ext_grf_steady",
]



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing miscellaneous tools.

Subpackages
^^^^^^^^^^^

.. currentmodule:: anaflow.tools

.. autosummary::
    laplace
    mean
    special
    coarse_graining

Functions
^^^^^^^^^

Annular mean
~~~~~~~~~~~~

.. currentmodule:: anaflow.tools.mean

Functions to calculate dimension dependent annular means of a function.

.. autosummary::
   annular_fmean
   annular_amean
   annular_gmean
   annular_hmean
   annular_pmean

Coarse Graining solutions
~~~~~~~~~~~~~~~~~~~~~~~~~

.. currentmodule:: anaflow.tools.coarse_graining

Effective Coarse Graining conductivity/transmissivity solutions.

.. autosummary::
   T_CG
   K_CG
   TPL_CG

Special
~~~~~~~

.. currentmodule:: anaflow.tools.special

Special functions.

.. autosummary::
   step_f
   specialrange
   specialrange_cut
   neuman2004_trans
   aniso

Laplace
~~~~~~~

.. currentmodule:: anaflow.tools.laplace

Helping functions related to the laplace-transformation

.. autosummary::
   get_lap
   get_lap_inv
"""
from anaflow.tools.mean import (
    annular_fmean,
    annular_amean,
    annular_gmean,
    annular_hmean,
    annular_pmean,
)
from anaflow.tools.special import (
    step_f,
    specialrange,
    specialrange_cut,
    neuman2004_trans,
    aniso,
)
from anaflow.tools.coarse_graining import T_CG, K_CG, TPL_CG
from anaflow.tools.laplace import get_lap, get_lap_inv

__all__ = [
    "get_lap",
    "get_lap_inv",
    "annular_fmean",
    "annular_amean",
    "annular_gmean",
    "annular_hmean",
    "annular_pmean",
    "step_f",
    "specialrange",
    "specialrange_cut",
    "neuman2004_trans",
    "aniso",
    "T_CG",
    "K_CG",
    "TPL_CG",
]



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as plt
from anaflow import theis, neuman2004


time_labels = ["10 s", "10 min", "10 h"]
time = [10, 600, 36000]      # 10s, 10min, 10h
rad = np.geomspace(0.05, 4)  # radius from the pumping well in [0, 4]
var = 0.5                    # variance of the log-transmissivity
len_scale = 10.0             # correlation length of the log-transmissivity
TG = 1e-4                    # the geometric mean of the transmissivity
TH = TG*np.exp(-var/2.0)     # the harmonic mean of the transmissivity
S = 1e-4                     # storativity
rate = -1e-4                   # pumping rate

head_TG = theis(time, rad, S, TG, rate)
head_TH = theis(time, rad, S, TH, rate)
head_ef = neuman2004(
    time=time,
    rad=rad,
    trans_gmean=TG,
    var=var,
    len_scale=len_scale,
    storage=S,
    rate=rate,
)
time_ticks = []
for i, step in enumerate(time):
    label_TG = "Theis($T_G$)" if i == 0 else None
    label_TH = "Theis($T_H$)" if i == 0 else None
    label_ef = "transient Neuman [2004]" if i == 0 else None
    plt.plot(rad, head_TG[i], label=label_TG, color="C"+str(i), linestyle="--")
    plt.plot(rad, head_TH[i], label=label_TH, color="C"+str(i), linestyle=":")
    plt.plot(rad, head_ef[i], label=label_ef, color="C"+str(i))
    time_ticks.append(head_ef[i][-1])

plt.xscale("log")
plt.xlabel("r in [m]")
plt.ylabel("h in [m]")
plt.legend()
ylim = plt.gca().get_ylim()
plt.gca().set_xlim([rad[0], rad[-1]])
ax2 = plt.gca().twinx()
ax2.set_yticks(time_ticks)
ax2.set_yticklabels(time_labels)
ax2.set_ylim(ylim)
plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
"""
Purpose
=======

Anaflow provides several analytical and semi-analytical solutions for the
groundwater-flow-equation.

Subpackages
===========

.. autosummary::
    flow
    tools

Solutions
=========

Homogeneous
^^^^^^^^^^^

.. currentmodule:: anaflow.flow.homogeneous

Solutions for homogeneous aquifers

.. autosummary::
   thiem
   theis
   grf

Heterogeneous
^^^^^^^^^^^^^

.. currentmodule:: anaflow.flow.heterogeneous

Solutions for heterogeneous aquifers

.. autosummary::
   ext_thiem_2d
   ext_thiem_3d
   ext_thiem_tpl
   ext_thiem_tpl_3d
   ext_theis_2d
   ext_theis_3d
   ext_theis_tpl
   ext_thiem_tpl_3d
   neuman2004
   neuman2004_steady

Extended GRF
^^^^^^^^^^^^

.. currentmodule:: anaflow.flow.ext_grf_model

The extended general radial flow model.

.. autosummary::
   ext_grf
   ext_grf_steady

Laplace
=======

.. currentmodule:: anaflow.tools.laplace

Helping functions related to the laplace-transformation

.. autosummary::
   get_lap
   get_lap_inv

Tools
=====

.. currentmodule:: anaflow.tools.special

Helping functions.

.. autosummary::
   step_f
   specialrange
   specialrange_cut


"""
from anaflow.flow import (
    thiem,
    theis,
    grf,
    ext_thiem_2d,
    ext_thiem_3d,
    ext_thiem_tpl,
    ext_thiem_tpl_3d,
    ext_theis_2d,
    ext_theis_3d,
    ext_theis_tpl,
    ext_theis_tpl_3d,
    neuman2004,
    neuman2004_steady,
    ext_grf,
    ext_grf_steady,
)
from anaflow.tools import (
    get_lap_inv,
    get_lap,
    step_f,
    specialrange,
    specialrange_cut,
)

try:
    from anaflow._version import __version__
except ImportError:  # pragma: nocover
    # package is not installed
    __version__ = "0.0.0.dev0"

__all__ = ["__version__"]
__all__ += [
    "thiem",
    "theis",
    "ext_thiem_2d",
    "ext_thiem_3d",
    "ext_thiem_tpl",
    "ext_thiem_tpl_3d",
    "ext_theis_2d",
    "ext_theis_3d",
    "ext_theis_tpl",
    "ext_theis_tpl_3d",
    "neuman2004",
    "neuman2004_steady",
    "grf",
    "ext_grf",
    "ext_grf_steady",
    "get_lap_inv",
    "get_lap",
    "step_f",
    "specialrange",
    "specialrange_cut",
]



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing special functions.

.. currentmodule:: anaflow.tools.special

The following functions are provided

.. autosummary::

   Shaper
   step_f
   sph_surf
   specialrange
   specialrange_cut
   aniso
   well_solution
   grf_solution
   inc_gamma
   tpl_hyp
   neuman2004_trans
"""

import numpy as np
from scipy.special import gamma, gammaincc, exp1, expn, hyp2f1

__all__ = [
    "Shaper",
    "step_f",
    "sph_surf",
    "specialrange",
    "specialrange_cut",
    "aniso",
    "well_solution",
    "grf_solution",
    "inc_gamma",
    "tpl_hyp",
    "neuman2004_trans",
]


class Shaper(object):
    """
    A class to reshape radius-time input-output in a good way.

    Parameters
    ----------
    time : :class:`numpy.ndarray` or :class:`float`, optional
        Array with all time-points where the function should be evaluated.
        Default: 0
    rad : :class:`numpy.ndarray` or :class:`float`, optional
        Array with all radii where the function should be evaluated.
        Default: 0
    struc_grid : :class:`bool`, optional
        If this is set to ``False``, the `rad` and `time` array will be merged
        and interpreted as single, r-t points. In this case they need to have
        the same shapes. Otherwise a structured t-r grid is created.
        Default: ``True``
    """

    def __init__(self, time=0, rad=0, struc_grid=True):
        self.time_scalar = np.isscalar(time)
        self.rad_scalar = np.isscalar(rad)
        self.time = np.array(time)
        self.rad = np.array(rad)
        self.struc_grid = struc_grid
        self.time_shape = self.time.shape
        self.rad_shape = self.rad.shape
        self.time = self.time.reshape(-1)
        self.rad = self.rad.reshape(-1)
        self.time_no = self.time.shape[0]
        self.rad_no = self.rad.shape[0]

        self.time_min = np.min(self.time)
        self.time_max = np.max(self.time)
        self.rad_min = np.min(self.rad)
        self.rad_max = np.max(self.rad)

        self.time_gz = self.time > 0
        self.time_mat = np.outer(
            self.time[self.time_gz], np.ones_like(self.rad)
        )
        self.rad_mat = np.outer(
            np.ones_like(self.time[self.time_gz]), self.rad
        )

        if not self.struc_grid and not self.rad_shape == self.time_shape:
            raise ValueError("No struc_grid: shape of time & radius differ")
        if np.any(self.time < 0.0):
            raise ValueError("The given times need to be positive.")
        if np.any(self.rad <= 0.0):
            raise ValueError("The given radii need to be non-negative.")

    def reshape(self, result):
        """Reshape a time-rad result according to the input shape."""
        np.asanyarray(result)
        if self.struc_grid:
            result = result.reshape(self.time_shape + self.rad_shape)
        elif self.rad_shape:
            result = np.diag(result).reshape(self.rad_shape)
        if self.time_scalar and self.rad_scalar:
            result = result.item()
        return result


def step_f(rad, r_part, f_part):
    """Callalbe step function."""
    return np.piecewise(
        np.array(rad),
        [
            np.logical_and(r1 <= rad, rad < r2)
            for r1, r2 in zip(r_part[:-1], r_part[1:])
        ],
        f_part,
    )


def sph_surf(dim):
    """Surface of the d-dimensional sphere."""
    return 2.0 * np.sqrt(np.pi) ** dim / gamma(dim / 2.0)


def specialrange(val_min, val_max, steps, typ="exp"):
    """
    Calculation of special point ranges.

    Parameters
    ----------
    val_min : :class:`float`
        Starting value.
    val_max : :class:`float`
        Ending value
    steps : :class:`int`
        Number of steps.
    typ : :class:`str` or :class:`float`, optional
        Setting the kind of range-distribution. One can choose between

        * ``"exp"``: for exponential behavior
        * ``"log"``: for logarithmic behavior
        * ``"geo"``: for geometric behavior
        * ``"lin"``: for linear behavior
        * ``"quad"``: for quadratic behavior
        * ``"cub"``: for cubic behavior
        * :class:`float`: here you can specifi any exponent ("quad" would be
          equivalent to 2)

        Default: ``"exp"``

    Returns
    -------
    :class:`numpy.ndarray`
        Array containing the special range

    Examples
    --------
    >>> specialrange(1,10,4)
    array([ 1.        ,  2.53034834,  5.23167968, 10.        ])
    """
    if typ in ["exponential", "exp"]:
        rng = np.expm1(
            np.linspace(np.log1p(val_min), np.log1p(val_max), steps)
        )
    elif typ in ["logarithmic", "log"]:
        rng = np.log(np.linspace(np.exp(val_min), np.exp(val_max), steps))
    elif typ in ["geometric", "geo", "geom"]:
        rng = np.geomspace(val_min, val_max, steps)
    elif typ in ["linear", "lin"]:
        rng = np.linspace(val_min, val_max, steps)
    elif typ in ["quadratic", "quad"]:
        rng = (np.linspace(np.sqrt(val_min), np.sqrt(val_max), steps)) ** 2
    elif typ in ["cubic", "cub"]:
        rng = (
            np.linspace(
                np.power(val_min, 1 / 3.0), np.power(val_max, 1 / 3.0), steps
            )
        ) ** 3
    elif isinstance(typ, (float, int)):
        rng = (
            np.linspace(
                np.power(val_min, 1.0 / typ),
                np.power(val_max, 1.0 / typ),
                steps,
            )
        ) ** typ
    else:
        print("specialrange: unknown typ '{}'. Using linear range".format(typ))
        rng = np.linspace(val_min, val_max, steps)

    return rng


def specialrange_cut(val_min, val_max, steps, val_cut=None, typ="exp"):
    """
    Calculation of special point ranges.

    Calculation of special point ranges with a cut-off value.

    Parameters
    ----------
    val_min : :class:`float`
        Starting value.
    val_max : :class:`float`
        Ending value
    steps : :class:`int`
        Number of steps.
    val_cut : :class:`float`, optional
        Cutting value, if val_max is bigger than this value, the last interval
        will be between val_cut and val_max.
        Default: 100.0
    typ : :class:`str` or :class:`float`, optional
        Setting the kind of range-distribution. One can choose between

        * ``"exp"``: for exponential behavior
        * ``"log"``: for logarithmic behavior
        * ``"geo"``: for geometric behavior
        * ``"lin"``: for linear behavior
        * ``"quad"``: for quadratic behavior
        * ``"cub"``: for cubic behavior
        * :class:`float`: here you can specifi any exponent ("quad" would be
          equivalent to 2)

        Default: ``"exp"``

    Returns
    -------
    :class:`numpy.ndarray`
        Array containing the special range

    Examples
    --------
    >>> specialrange_cut(1,10,4)
    array([ 1.        ,  2.53034834,  5.23167968, 10.        ])
    """
    val_cut = 100.0 if val_cut is None else val_cut
    if val_max > val_cut:
        rng = specialrange(val_min, val_cut, steps - 1, typ)
        return np.hstack((rng, val_max))
    return specialrange(val_min, val_max, steps, typ)


def aniso(e):
    """
    The anisotropy function.

    Known from ''Dagan (1989)''[R2]_.

    Parameters
    ----------
    e : :class:`float`
        Anisotropy-ratio of the vertical and horizontal corralation-lengths

    Returns
    -------
    aniso : :class:`float`
        Value of the anisotropy function for the given value.

    Raises
    ------
    ValueError
        If the Anisotropy-ratio ``e`` is not within 0 and 1.


    References
    ----------
    .. [R2] Dagan, G., ''Flow and Transport on Porous Formations'',
           Springer Verlag, New York, 1989.

    Examples
    --------
    >>> aniso(0.5)
    0.2363998587187151
    """
    if e < 0 or e > 1:
        raise ValueError("Anisotropy ratio 'e' must be within 0 and 1")

    if np.isclose(e, 1):
        res = 1.0 / 3.0
    elif np.isclose(e, 0):
        res = 0.0
    else:
        res = e / (2 * (1.0 - e ** 2))
        res *= (
            1.0
            / np.sqrt(1.0 - e ** 2)
            * np.arctan(np.sqrt(1.0 / e ** 2 - 1.0))
            - e
        )

    return res


def well_solution(
    time,
    rad,
    storage,
    transmissivity,
    rate=-1e-4,
    h_bound=0.0,
    struc_grid=True,
):
    """
    The classical Theis solution.

    The classical Theis solution for transient flow under a pumping condition
    in a confined and homogeneous aquifer.

    This solution was presented in ''Theis 1935''[R9]_.

    Parameters
    ----------
    time : :class:`numpy.ndarray`
        Array with all time-points where the function should be evaluated.
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated.
    storage : :class:`float`
        Storage of the aquifer.
    transmissivity : :class:`float`
        Transmissivity of the aquifer.
    rate : :class:`float`, optional
        Pumpingrate at the well. Default: -1e-4
    h_bound : :class:`float`, optional
        Reference head at the outer boundary at infinity. Default: ``0.0``
    struc_grid : :class:`bool`, optional
        If this is set to "False", the "rad" and "time" array will be merged
        and interpreted as single, r-t points. In this case they need to have
        the same shapes. Otherwise a structured r-t grid is created.
        Default: ``True``

    Returns
    -------
    head : :class:`numpy.ndarray`
        Array with all heads at the given radii and time-points.

    Raises
    ------
    ValueError
        If ``rad`` is not positiv.
    ValueError
        If ``time`` is negative.
    ValueError
        If shape of ``rad`` and ``time`` differ in case of
        ``struc_grid`` is ``True``.
    ValueError
        If ``transmissivity`` is not positiv.
    ValueError
        If ``storage`` is not positiv.

    References
    ----------
    .. [R9] Theis, C.,
       ''The relation between the lowering of the piezometric surface and the
       rate and duration of discharge of a well using groundwater storage'',
       Trans. Am. Geophys. Union, 16, 519-524, 1935

    Notes
    -----
    The parameters ``rad``, ``T`` and ``S`` will be checked for positivity.
    If you want to use cartesian coordiantes, just use the formula
    ``r = sqrt(x**2 + y**2)``

    Examples
    --------
    >>> well_solution([10,100], [1,2,3], 0.001, 0.001, -0.001)
    array([[-0.24959541, -0.14506368, -0.08971485],
           [-0.43105106, -0.32132823, -0.25778313]])
    """
    Input = Shaper(time, rad, struc_grid)

    if not transmissivity > 0.0:
        raise ValueError("The Transmissivity needs to be positive.")
    if not storage > 0.0:
        raise ValueError("The Storage needs to be positive.")

    time_mat = Input.time_mat
    rad_mat = Input.rad_mat

    res = np.zeros((Input.time_no, Input.rad_no))
    res[Input.time_gz, :] = (
        rate
        / (4.0 * np.pi * transmissivity)
        * exp1(rad_mat ** 2 * storage / (4 * transmissivity * time_mat))
    )
    res = Input.reshape(res)
    if rate > 0:
        res = np.maximum(res, 0)
    else:
        res = np.minimum(res, 0)
    # add the reference head
    res += h_bound
    return res


def grf_solution(
    time,
    rad,
    storage,
    conductivity,
    dim=2,
    lat_ext=1.0,
    rate=-1e-4,
    h_bound=0.0,
    struc_grid=True,
):
    """
    The general radial flow (GRF) model for a pumping test.

    Parameters
    ----------
    time : :class:`numpy.ndarray`
        Array with all time-points where the function should be evaluated.
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated.
    storage : :class:`float`
        Storage coefficient of the aquifer.
    conductivity : :class:`float`
        Conductivity of the aquifer.
    dim : :class:`float`, optional
        Fractional dimension of the aquifer. Default: ``2.0``
    lat_ext : :class:`float`, optional
        Lateral extend of the aquifer. Default: ``1.0``
    rate : :class:`float`, optional
        Pumpingrate at the well. Default: -1e-4
    h_bound : :class:`float`, optional
        Reference head at the outer boundary at infinity. Default: ``0.0``
    struc_grid : :class:`bool`, optional
        If this is set to "False", the "rad" and "time" array will be merged
        and interpreted as single, r-t points. In this case they need to have
        the same shapes. Otherwise a structured r-t grid is created.
        Default: ``True``

    Returns
    -------
    head : :class:`numpy.ndarray`
        Array with all heads at the given radii and time-points.

    Raises
    ------
    ValueError
        If ``rad`` is not positiv.
    ValueError
        If ``time`` is negative.
    ValueError
        If shape of ``rad`` and ``time`` differ in case of
        ``struc_grid`` is ``True``.
    ValueError
        If ``conductivity`` is not positiv.
    ValueError
        If ``storage`` is not positiv.
    """
    Input = Shaper(time, rad, struc_grid)

    if not conductivity > 0.0:
        raise ValueError("The Conductivity needs to be positive.")
    if not storage > 0.0:
        raise ValueError("The Storage needs to be positive.")

    time_mat = Input.time_mat
    rad_mat = Input.rad_mat
    u = storage * rad_mat ** 2 / (4 * conductivity * time_mat)
    nu = 1.0 - dim / 2.0

    res = np.zeros((Input.time_no, Input.rad_no))
    res[Input.time_gz, :] = inc_gamma(-nu, u)
    res[Input.time_gz, :] *= (
        rate
        / (4.0 * np.pi ** (1 - nu) * conductivity * lat_ext ** (3.0 - dim))
        * rad_mat ** (2 * nu)
    )
    res = Input.reshape(res)
    if rate > 0:
        res = np.maximum(res, 0)
    else:
        res = np.minimum(res, 0)
    # add the reference head
    res += h_bound
    return res


def inc_gamma(s, x):
    r"""The (upper) incomplete gamma function.

    Given by: :math:`\Gamma(s,x) = \int_x^{\infty} t^{s-1}\,e^{-t}\,{\rm d}t`

    Parameters
    ----------
    s : :class:`float`
        exponent in the integral
    x : :class:`numpy.ndarray`
        input values
    """
    if np.isclose(s, 0):
        return exp1(x)
    if np.isclose(s, np.around(s)) and s < -0.5:
        return x ** (s - 1) * expn(int(1 - np.around(s)), x)
    if s < 0:
        return (inc_gamma(s + 1, x) - x ** s * np.exp(-x)) / s
    return gamma(s) * gammaincc(s, x)


def tpl_hyp(rad, dim, hurst, corr, prop):
    """Hyp_2F1 for the TPL CG model."""
    x = 1.0 / (1.0 + (prop * rad / corr) ** 2)
    return x ** (dim / 2.0) * hyp2f1(dim / 2.0, 1, dim / 2.0 + 1 + hurst, x)


def neuman2004_trans(rad, trans_gmean, var, len_scale):
    r"""The apparent transmissivity from Neuman 2004.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    trans_gmean : :class:`float`
        Geometric-mean transmissivity.
    var : :class:`float`
        Variance of log-transmissivity.
    len_scale : :class:`float`
        Correlation-length of log-transmissivity.
    """
    t_h = trans_gmean * np.exp(-var / 2)
    return t_h + (trans_gmean - t_h) * neuman2004_phi(0.5 * rad / len_scale)


def neuman2004_phi(alpha):
    r"""The phi function from Neuman 2004.

    Parameters
    ----------
    alpha : :class:`numpy.ndarray`
        The ratio r/(2l)
    """
    alpha = np.array(np.abs(alpha), dtype=np.double)
    a_lo = alpha < 1
    res = np.ones_like(alpha)
    res[a_lo] = 3 * alpha[a_lo] ** 2 - 2 * alpha[a_lo] ** 3
    return res


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing helper functions related to coarse graining.

.. currentmodule:: anaflow.tools.coarse_graining

The following functions are provided

.. autosummary::

   T_CG
   T_CG_inverse
   T_CG_error
   K_CG
   K_CG_inverse
   K_CG_error
   TPL_CG
   TPL_CG_error
"""
# pylint: disable=C0103
import numpy as np
from scipy.optimize import root

from anaflow.tools.special import aniso, tpl_hyp

__all__ = [
    "T_CG",
    "T_CG_inverse",
    "T_CG_error",
    "K_CG",
    "K_CG_inverse",
    "K_CG_error",
    "TPL_CG",
    "TPL_CG_error",
]


def T_CG(rad, trans_gmean, var, len_scale, T_well=None, prop=1.6):
    """
    The coarse-graining Transmissivity.

    This solution was presented in ''Schneider & Attinger 2008''[R3]_.

    This functions gives an effective
    transmissivity for 2D pumpingtests in heterogenous aquifers, where the
    transmissivity is following a log-normal distribution and a gaussian
    correlation function.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    trans_gmean : :class:`float`
        Geometric-mean transmissivity.
    var : :class:`float`
        Variance of log-transmissivity.
    len_scale : :class:`float`
        Correlation-length of log-transmissivity.
    T_well : :class:`float`, optional
        Explicit transmissivity value at the well. Harmonic mean by default.
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``

    Returns
    -------
    T_CG : :class:`numpy.ndarray`
        Array containing the effective transmissivity values.

    References
    ----------
    .. [R3] Schneider, C. and Attinger, S.,
       ''Beyond thiem: A new method for interpreting large scale
       pumping tests in heterogeneous aquifers.''
       Water resources research, 44(4), 2008

    Examples
    --------
    >>> T_CG([1,2,3], 0.001, 1, 10, 2)
    array([0.00061831, 0.00064984, 0.00069236])
    """
    chi = -var / 2.0 if T_well is None else np.log(T_well / trans_gmean)
    return trans_gmean * np.exp(chi / (1.0 + (prop * rad / len_scale) ** 2))


def T_CG_inverse(T, trans_gmean, var, len_scale, T_well=None, prop=1.6):
    """
    The inverse coarse-graining Transmissivity.

    See: :func:`T_CG`

    Parameters
    ----------
    T : :class:`numpy.ndarray`
        Array with all transmissivity values
        where the function should be evaluated
    trans_gmean : :class:`float`
        Geometric-mean transmissivity.
    var : :class:`float`
        Variance of log-transmissivity.
    len_scale : :class:`float`
        Correlation-length of log-transmissivity.
    T_well : :class:`float`, optional
        Explicit transmissivity value at the well. Harmonic mean by default.
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``

    Returns
    -------
    rad : :class:`numpy.ndarray`
        Array containing the radii belonging to the given transmissivity values

    Examples
    --------
    >>> T_CG_inverse([7e-4,8e-4,9e-4], 0.001, 1, 10, 2)
    array([3.16952925, 5.56935826, 9.67679026])
    """
    chi = -var / 2.0 if T_well is None else np.log(T_well / trans_gmean)
    return (len_scale / prop) * np.sqrt(chi / np.log(T / trans_gmean) - 1.0)


def T_CG_error(err, trans_gmean, var, len_scale, T_well=None, prop=1.6):
    """
    Calculating the radial-point for given error.

    Calculating the radial-point where the relative error of the farfield
    value is less than the given tollerance.
    See: :func:`T_CG`

    Parameters
    ----------
    err : :class:`float`
        Given relative error for the farfield transmissivity
    trans_gmean : :class:`float`
        Geometric-mean transmissivity.
    var : :class:`float`
        Variance of log-transmissivity.
    len_scale : :class:`float`
        Correlation-length of log-transmissivity.
    T_well : :class:`float`, optional
        Explicit transmissivity value at the well. Harmonic mean by default.
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``

    Returns
    -------
    rad : :class:`float`
        Radial point, where the relative error is less than the given one.

    Examples
    --------
    >>> T_CG_error(0.01, 0.001, 1, 10, 2)
    34.91045016779039
    """
    chi = -var / 2.0 if T_well is None else np.log(T_well / trans_gmean)
    if chi > 0.0:
        if chi / np.log(1.0 + err) >= 1.0:
            return (len_scale / prop) * np.sqrt(chi / np.log(1.0 + err) - 1.0)
        # standard value if the error is less then the variation
        return 0
    if chi / np.log(1.0 - err) >= 1.0:
        return (len_scale / prop) * np.sqrt(chi / np.log(1.0 - err) - 1.0)
    # standard value if the error is less then the variation
    return 0


def K_CG(rad, cond_gmean, var, len_scale, anis, K_well="KH", prop=1.6):
    """
    The coarse-graining conductivity.

    This solution was presented in ''Zech 2013''[R8]_.

    This functions gives an effective
    conductivity for 3D pumpingtests in heterogenous aquifers, where the
    conductivity is following a log-normal distribution and a gaussian
    correlation function and taking vertical anisotropy into account.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    cond_gmean : :class:`float`
        Geometric-mean conductivity.
    var : :class:`float`
        Variance of the log-conductivity.
    len_scale : :class:`float`
        Corralation-length of log-conductivity.
    anis : :class:`float`
        Anisotropy-ratio of the vertical and horizontal corralation-lengths.
    K_well : string/float, optional
        Explicit conductivity value at the well. One can choose between the
        harmonic mean (``"KH"``), the arithmetic mean (``"KA"``) or an
        arbitrary float value. Default: ``"KH"``
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``

    Returns
    -------
    K_CG : :class:`numpy.ndarray`
        Array containing the effective conductivity values.

    References
    ----------
    .. [R8] Zech, A.
       ''Impact of Aqifer Heterogeneity on Subsurface Flow and Salt
       Transport at Different Scales: from a method determine parameters
       of heterogeneous permeability at local scale to a large-scale model
       for the sedimentary basin of Thuringia.''
       PhD thesis, Friedrich-Schiller-Universität Jena, 2013

    Examples
    --------
    >>> K_CG([1,2,3], 0.001, 1, 10, 1, 2)
    array([0.00063008, 0.00069285, 0.00077595])
    """
    K_efu = cond_gmean * np.exp(var * (0.5 - aniso(anis)))
    if K_well == "KH":
        chi = var * (aniso(anis) - 1.0)
    elif K_well == "KA":
        chi = var * aniso(anis)
    else:
        chi = np.log(K_well / K_efu)

    return K_efu * np.exp(
        chi
        / np.sqrt(1.0 + (prop * rad / (len_scale * anis ** (1.0 / 3.0))) ** 2)
        ** 3
    )


def K_CG_inverse(K, cond_gmean, var, len_scale, anis, K_well="KH", prop=1.6):
    """
    The inverse coarse-graining conductivity.

    See: :func:`K_CG`

    Parameters
    ----------
    K : :class:`numpy.ndarray`
        Array with all conductivity values
        where the function should be evaluated
    cond_gmean : :class:`float`
        Geometric-mean conductivity.
    var : :class:`float`
        Variance of the log-conductivity.
    len_scale : :class:`float`
        Corralation-length of log-conductivity.
    anis : :class:`float`
        Anisotropy-ratio of the vertical and horizontal corralation-lengths.
    K_well : string/float, optional
        Explicit conductivity value at the well. One can choose between the
        harmonic mean (``"KH"``), the arithmetic mean (``"KA"``) or an
        arbitrary float value. Default: ``"KH"``
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``

    Returns
    -------
    rad : :class:`numpy.ndarray`
        Array containing the radii belonging to the given conductivity values

    Examples
    --------
    >>> K_CG_inverse([7e-4,8e-4,9e-4], 0.001, 1, 10, 1, 2)
    array([2.09236867, 3.27914996, 4.52143956])
    """
    K_efu = cond_gmean * np.exp(var * (0.5 - aniso(anis)))
    if K_well == "KH":
        chi = var * (aniso(anis) - 1.0)
    elif K_well == "KA":
        chi = var * aniso(anis)
    else:
        chi = np.log(K_well / K_efu)

    return (
        len_scale
        * anis ** (1.0 / 3.0)
        / prop
        * np.sqrt((chi / np.log(K / K_efu)) ** (2.0 / 3.0) - 1.0)
    )


def K_CG_error(err, cond_gmean, var, len_scale, anis, K_well="KH", prop=1.6):
    """
    Calculating the radial-point for given error.

    Calculating the radial-point where the relative error of the farfield
    value is less than the given tollerance.
    See: :func:`K_CG`

    Parameters
    ----------
    err : :class:`float`
        Given relative error for the farfield conductivity
    cond_gmean : :class:`float`
        Geometric-mean conductivity.
    var : :class:`float`
        Variance of the log-conductivity.
    len_scale : :class:`float`
        Corralation-length of log-conductivity.
    anis : :class:`float`
        Anisotropy-ratio of the vertical and horizontal corralation-lengths.
    K_well : string/float, optional
        Explicit conductivity value at the well. One can choose between the
        harmonic mean (``"KH"``), the arithmetic mean (``"KA"``) or an
        arbitrary float value. Default: ``"KH"``
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``

    Returns
    -------
    rad : :class:`float`
        Radial point, where the relative error is less than the given one.

    Examples
    --------
    >>> K_CG_error(0.01, 0.001, 1, 10, 1, 2)
    19.612796453639845
    """
    K_efu = cond_gmean * np.exp(var * (0.5 - aniso(anis)))
    if K_well == "KH":
        chi = var * (aniso(anis) - 1.0)
    elif K_well == "KA":
        chi = var * aniso(anis)
    else:
        chi = np.log(K_well / K_efu)

    coef = len_scale * anis ** (1.0 / 3.0) / prop

    if chi > 0.0:
        if chi / np.log(1.0 + err) >= 1.0:
            return coef * np.sqrt(
                (chi / np.log(1.0 + err)) ** (2.0 / 3.0) - 1.0
            )
        # standard value if the error is less then the variation
        return 0

    if chi / np.log(1.0 - err) >= 1.0:
        return coef * np.sqrt((chi / np.log(1.0 - err)) ** (2.0 / 3.0) - 1.0)
    # standard value if the error is less then the variation
    return 0


def TPL_CG(
    rad,
    cond_gmean,
    len_scale,
    hurst,
    var=None,
    c=1.0,
    anis=1,
    dim=2.0,
    K_well="KH",
    prop=1.6,
):
    """
    The gaussian truncated power-law coarse-graining conductivity.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    cond_gmean : :class:`float`
        Geometric-mean conductivity
    len_scale : :class:`float`
        upper bound of the corralation-length of conductivity-distribution
    hurst: :class:`float`
        Hurst coefficient of the TPL model. Should be in (0, 1).
    var : :class:`float` or :any:`None`, optional
        Variance of log-conductivity
        If given, c will be calculated accordingly.
        Default: :any:`None`
    c : :class:`float`, optional
        Intensity of variation in the TPL model.
        Is overwritten if var is given.
        Default: ``1.0``
    anis : :class:`float`, optional
        Anisotropy-ratio of the vertical and horizontal corralation-lengths.
        This is only applied in 3 dimensions.
        Default: 1.0
    dim: :class:`float`, optional
        Dimension of space.
        Default: ``2.0``
    K_well :  :class:`str` or  :class:`float`, optional
        Explicit conductivity value at the well. One can choose between the
        harmonic mean (``"KH"``),
        the arithmetic mean (``"KA"``) or an arbitrary float
        value. Default: ``"KH"``
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``

    Returns
    -------
    TPL_CG : :class:`numpy.ndarray`
        Array containing the effective conductivity values.
    """
    # handle special case in 3D with anisotropy
    anis = 1.0 if not np.isclose(dim, 3) else anis
    ani = aniso(anis) if np.isclose(dim, 3) else 1.0 / dim
    var = c * len_scale ** (2 * hurst) / (2 * hurst) if var is None else var
    K_efu = cond_gmean * np.exp(var * (0.5 - ani))
    if K_well == "KH":
        chi = var * (ani - 1.0)
    elif K_well == "KA":
        chi = var * ani
    else:
        chi = np.log(K_well / K_efu)

    return K_efu * np.exp(
        (chi * 2.0 * hurst / (dim + 2.0 * hurst))
        * tpl_hyp(rad, dim, hurst, len_scale * anis ** (1 / 3.0), prop)
    )


def TPL_CG_error(
    err,
    cond_gmean,
    len_scale,
    hurst,
    var=None,
    c=1.0,
    anis=1,
    dim=2.0,
    K_well="KH",
    prop=1.6,
):
    """
    Calculating the radial-point for given error.

    Calculating the radial-point where the relative error of the farfield
    value is less than the given tollerance.
    See: :func:`TPL_CG`

    Parameters
    ----------
    err : :class:`float`
        Given relative error for the farfield conductivity
    cond_gmean : :class:`float`
        Geometric-mean conductivity
    len_scale : :class:`float`
        upper bound of the corralation-length of conductivity-distribution
    hurst: :class:`float`
        Hurst coefficient of the TPL model. Should be in (0, 1).
    var : :class:`float` or :any:`None`, optional
        Variance of log-conductivity
        If given, c will be calculated accordingly.
        Default: :any:`None`
    c : :class:`float`, optional
        Intensity of variation in the TPL model.
        Is overwritten if var is given.
        Default: ``1.0``
    anis : :class:`float`, optional
        Anisotropy-ratio of the vertical and horizontal corralation-lengths.
        This is only applied in 3 dimensions.
        Default: 1.0
    dim: :class:`float`, optional
        Dimension of space.
        Default: ``2.0``
    K_well :  :class:`str` or  :class:`float`, optional
        Explicit conductivity value at the well. One can choose between the
        harmonic mean (``"KH"``),
        the arithmetic mean (``"KA"``) or an arbitrary float
        value. Default: ``"KH"``
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``

    Returns
    -------
    rad : :class:`float`
        Radial point, where the relative error is less than the given one.
    """
    # handle special case in 3D with anisotropy
    anis = 1.0 if not np.isclose(dim, 3) else anis
    ani = aniso(anis) if np.isclose(dim, 3) else 1.0 / dim
    var = c * len_scale ** (2 * hurst) / (2 * hurst) if var is None else var
    K_efu = cond_gmean * np.exp(var * (0.5 - ani))
    if K_well == "KH":
        chi = var * (ani - 1.0)
    elif K_well == "KA":
        chi = var * ani
    else:
        chi = np.log(K_well / K_efu)
    Kw = np.exp(chi + np.log(K_efu))

    # define a curve, that has its root at the wanted percentile
    if chi > 0:
        per = (1 + err) * K_efu
        if not per < Kw:
            return 0
    elif chi < 0:
        per = (1 - err) * K_efu
        if not per > Kw:
            return 0
    else:
        return 0

    def curve(x):
        """Curve for fitting."""
        return (
            TPL_CG(
                x,
                cond_gmean=cond_gmean,
                len_scale=len_scale,
                hurst=hurst,
                var=var,
                c=c,
                anis=anis,
                dim=dim,
                K_well=K_well,
                prop=prop,
            )
            - per
        )

    return root(curve, 2 * len_scale)["x"][0]


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as plt
from anaflow import ext_theis_2d, neuman2004


time_labels = ["10 s", "10 min", "10 h"]
time = [10, 600, 36000]      # 10s, 10min, 10h
rad = np.geomspace(0.05, 4)  # radius from the pumping well in [0, 4]
TG = 1e-4                    # the geometric mean of the transmissivity
var = 0.5                    # correlation length of the log-transmissivity
len_scale = 10.0             # variance of the log-transmissivity
S = 1e-4                     # storativity
rate = -1e-4                 # pumping rate

head1 = ext_theis_2d(time, rad, S, TG, var, len_scale, rate)
head2 = neuman2004(time, rad, S, TG, var, len_scale, rate)
time_ticks=[]
for i, step in enumerate(time):
    label1 = "extended Theis 2D" if i == 0 else None
    label2 = "Transient Neuman 2004" if i == 0 else None
    plt.plot(rad, head1[i], label=label1, color="C"+str(i))
    plt.plot(rad, head2[i], label=label2, color="C"+str(i), linestyle="--")
    time_ticks.append(head1[i][-1])

plt.title("$T_G={}$, $\sigma^2={}$, $\ell={}$, $S={}$".format(TG, var, len_scale, S))
plt.xlabel("r in [m]")
plt.ylabel("h in [m]")
plt.legend()
ylim = plt.gca().get_ylim()
plt.gca().set_xlim([0, rad[-1]])
ax2 = plt.gca().twinx()
ax2.set_yticks(time_ticks)
ax2.set_yticklabels(time_labels)
ax2.set_ylim(ylim)
plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing several mean calculating routines.

.. currentmodule:: anaflow.tools.mean

The following functions are provided

.. autosummary::

   annular_fmean
   annular_amean
   annular_gmean
   annular_hmean
   annular_pmean
"""
# pylint: disable=E1137, C0103
import numpy as np

from scipy.integrate import quad as integ

__all__ = [
    "annular_fmean",
    "annular_amean",
    "annular_gmean",
    "annular_hmean",
    "annular_pmean",
]


def annular_fmean(
    func, val_arr, f_def, f_inv, ann_dim=2, arg_dict=None, **kwargs
):
    r"""
    Calculating the annular generalized f-mean.

    Calculating the annular generalized f-mean of a radial symmetric function
    for given consecutive radii defining annuli by the following formula

    .. math::
       \varphi_i=f^{-1}\left(\frac{d}{r_{i+1}^d-r_i^d}
       \intop^{r_{i+1}}_{r_i} r^{d-1}\cdot f(\varphi(r))\, dr \right)

    Parameters
    ----------
    func : :any:`callable`
        Function that should be used (:math:`\varphi` in the formula).
        The first argument needs to be the radial variable:
        ``func(r, **kwargs)``
    val_arr : :class:`numpy.ndarray`
        Radii defining the annuli.
    ann_dim : :class:`float`, optional
        The dimension of the annuli.
        Default: ``2.0``
    f_def : :any:`callable`
        Function defining the f-mean.
    f_inv : :any:`callable`
        Inverse of the function defining the f-mean.
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``.
        This is designed for overlapping keywords in ``annular_fmean`` and
        ``func``.
        Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all calculated arithmetic means

    Raises
    ------
    ValueError
        If `func` is not callable.
    ValueError
        If `f_def` is not callable.
    ValueError
        If `f_inv` is not callable.
    ValueError
        If ``val_arr`` has less than 2 values.
    ValueError
        If ``val_arr`` is not sorted in incresing order.

    Notes
    -----
    If the last value in val_arr is "inf", the given function should provide
    a value for "inf" as input: ``func(float("inf"))``
    """
    arg_dict = {} if arg_dict is None else arg_dict
    kwargs.update(arg_dict)

    val_arr = np.array(val_arr, dtype=np.double).reshape(-1)
    parts = len(val_arr) - 1

    if not callable(func):
        raise ValueError("The given function needs to be callable")
    if not callable(f_def):
        raise ValueError("The f-mean function needs to be callable")
    if not callable(f_inv):
        raise ValueError("The inverse f-mean function needs to be callable")
    if not np.all(
        np.isclose(
            f_inv(f_def(func(val_arr, **kwargs))), func(val_arr, **kwargs)
        )
    ):
        raise ValueError("f_def and f_inv need to be inverse to each other")
    if len(val_arr) < 2:
        raise ValueError("To few input values in val_arr. Need at least 2.")
    if not all(val_arr[i] < val_arr[i + 1] for i in range(len(val_arr) - 1)):
        raise ValueError("The input values need to be sorted")

    def integrand(val):
        """Integrand for the f-mean ``r^(d-1)*f(phi(r))``."""
        return val ** (ann_dim - 1) * f_def(func(val, **kwargs))

    # creating the output array
    func_arr = np.zeros_like(val_arr[:-1], dtype=np.double)

    # iterating over the input values
    for i in range(parts):
        # if one side is infinity, the function is evaluated at infinity
        if val_arr[i + 1] == np.inf:
            func_arr[i] = func(np.inf, **kwargs)
        else:
            func_arr[i] = f_inv(
                ann_dim
                * integ(integrand, val_arr[i], val_arr[i + 1])[0]
                / (val_arr[i + 1] ** ann_dim - val_arr[i] ** ann_dim)
            )

    return func_arr


def annular_amean(func, val_arr, ann_dim=2, arg_dict=None, **kwargs):
    r"""
    Calculating the annular arithmetic mean.

    Calculating the annular arithmetic mean of a radial symmetric function
    for given consecutive radii defining annuli by the following formula

    .. math::
       \varphi_i=\frac{d}{r_{i+1}^d-r_i^d}
       \intop^{r_{i+1}}_{r_i} r^{d-1}\cdot \varphi(r)\, dr

    Parameters
    ----------
    func : :any:`callable`
        Function that should be used (:math:`\varphi` in the formula).
        The first argument needs to be the radial variable:
        ``func(r, **kwargs)``
    val_arr : :class:`numpy.ndarray`
        Radii defining the annuli.
    ann_dim : :class:`float`, optional
        The dimension of the annuli.
        Default: ``2.0``
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``.
        This is designed for overlapping keywords in ``annular_amean`` and
        ``func``.
        Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all calculated arithmetic means

    Raises
    ------
    ValueError
        If `func` is not callable.
    ValueError
        If ``val_arr`` has less than 2 values.
    ValueError
        If ``val_arr`` is not sorted in incresing order.

    Notes
    -----
    If the last value in val_arr is "inf", the given function should provide
    a value for "inf" as input: ``func(float("inf"))``
    """
    return annular_fmean(
        func=func,
        val_arr=val_arr,
        f_def=lambda x: x,
        f_inv=lambda x: x,
        ann_dim=ann_dim,
        arg_dict=arg_dict,
        **kwargs
    )


def annular_gmean(func, val_arr, ann_dim=2, arg_dict=None, **kwargs):
    r"""
    Calculating the annular geometric mean.

    Calculating the annular geometric mean of a radial symmetric function
    for given consecutive radii defining annuli by the following formula

    .. math::
       \varphi_i=\exp\left(\frac{d}{r_{i+1}^d-r_i^d}
       \intop^{r_{i+1}}_{r_i} r^{d-1}\cdot \ln(\varphi(r))\, dr \right)

    Parameters
    ----------
    func : :any:`callable`
        Function that should be used (:math:`\varphi` in the formula).
        The first argument needs to be the radial variable:
        ``func(r, **kwargs)``
    val_arr : :class:`numpy.ndarray`
        Radii defining the annuli.
    ann_dim : :class:`float`, optional
        The dimension of the annuli.
        Default: ``2.0``
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``.
        This is designed for overlapping keywords in ``annular_gmean`` and
        ``func``.
        Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all calculated geometric means

    Raises
    ------
    ValueError
        If `func` is not callable.
    ValueError
        If ``val_arr`` has less than 2 values.
    ValueError
        If ``val_arr`` is not sorted in incresing order.

    Notes
    -----
    If the last value in val_arr is "inf", the given function should provide
    a value for "inf" as input: ``func(float("inf"))``

    Examples
    --------
    >>> f = lambda x: x**2
    >>> annular_gmean(f, [1,2,3])
    array([ 2.33588885,  6.33423311])
    """
    return annular_fmean(
        func=func,
        val_arr=val_arr,
        f_def=np.log,
        f_inv=np.exp,
        ann_dim=ann_dim,
        arg_dict=arg_dict,
        **kwargs
    )


def annular_hmean(func, val_arr, ann_dim=2, arg_dict=None, **kwargs):
    r"""
    Calculating the annular harmonic mean.

    Calculating the annular harmonic mean of a radial symmetric function
    for given consecutive radii defining annuli by the following formula

    .. math::
       \varphi_i=\left(\frac{d}{r_{i+1}^d-r_i^d}
       \intop^{r_{i+1}}_{r_i} r^{d-1}\cdot \varphi(r)^{-1}\, dr \right)^{-1}

    Parameters
    ----------
    func : :any:`callable`
        Function that should be used (:math:`\varphi` in the formula).
        The first argument needs to be the radial variable:
        ``func(r, **kwargs)``
    val_arr : :class:`numpy.ndarray`
        Radii defining the annuli.
    ann_dim : :class:`float`, optional
        The dimension of the annuli.
        Default: ``2.0``
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``.
        This is designed for overlapping keywords in ``annular_hmean`` and
        ``func``.
        Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all calculated geometric means

    Raises
    ------
    ValueError
        If `func` is not callable.
    ValueError
        If ``val_arr`` has less than 2 values.
    ValueError
        If ``val_arr`` is not sorted in incresing order.

    Notes
    -----
    If the last value in val_arr is "inf", the given function should provide
    a value for "inf" as input: ``func(float("inf"))``
    """
    return annular_fmean(
        func=func,
        val_arr=val_arr,
        f_def=lambda x: 1.0 / x,
        f_inv=lambda x: 1.0 / x,
        ann_dim=ann_dim,
        arg_dict=arg_dict,
        **kwargs
    )


def annular_pmean(func, val_arr, p=2.0, ann_dim=2, arg_dict=None, **kwargs):
    r"""
    Calculating the annular p-mean.

    Calculating the annular p-mean of a radial symmetric function
    for given consecutive radii defining annuli by the following formula

    .. math::
       \varphi_i=\left(\frac{d}{r_{i+1}^d-r_i^d}
       \intop^{r_{i+1}}_{r_i} r^{d-1}\cdot\varphi(r)^p\, dr
       \right)^{\frac{1}{p}}

    Parameters
    ----------
    func : :any:`callable`
        Function that should be used (:math:`\varphi` in the formula).
        The first argument needs to be the radial variable:
        ``func(r, **kwargs)``
    val_arr : :class:`numpy.ndarray`
        Radii defining the annuli.
    p : :class:`float`, optional
        The potency defining the p-mean.
        Default: ``2.0``
    ann_dim : :class:`float`, optional
        The dimension of the annuli.
        Default: ``2.0``
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``.
        This is designed for overlapping keywords in ``annular_pmean`` and
        ``func``.
        Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all calculated p-means

    Raises
    ------
    ValueError
        If `func` is not callable.
    ValueError
        If ``val_arr`` has less than 2 values.
    ValueError
        If ``val_arr`` is not sorted in incresing order.

    Notes
    -----
    If the last value in val_arr is "inf", the given function should provide
    a value for "inf" as input: ``func(float("inf"))``
    """
    # if p is 0, the limit-case of the geometric mean is returned
    if np.isclose(p, 0):
        return annular_gmean(func, val_arr, ann_dim, arg_dict, **kwargs)

    return annular_fmean(
        func=func,
        val_arr=val_arr,
        f_def=lambda x: np.power(x, p),
        f_inv=lambda x: np.power(x, 1.0 / p),
        ann_dim=ann_dim,
        arg_dict=arg_dict,
        **kwargs
    )



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as plt
from anaflow import ext_grf, ext_grf_steady, grf

time = 1e4                   # time point for steady state
rad = np.geomspace(0.1, 10)  # radius from the pumping well in [0, 4]
r_ref = 10.0                 # reference radius
K = 1e-4                     # the geometric mean of the transmissivity
rate = -1e-4                 # pumping rate
dim = 1.5                    # using a fractional dimension

head1 = ext_grf_steady(rad, r_ref, K, dim=dim, rate=rate)
head2 = ext_grf(time, rad, [1e-4], [K], [0, r_ref], dim=dim, rate=rate)
head3 = grf(time, rad, 1e-4, K, dim=dim, rate=rate)
head3 -= head3[-1]  # quasi-steady

plt.plot(rad, head1, label="Ext GRF steady")
plt.plot(rad, head2, label="Ext GRF (t={})".format(time), linestyle="--")
plt.plot(rad, head3, label="GRF quasi-steady (t={})".format(time), linestyle=":")

plt.xlabel("r in [m]")
plt.ylabel("h in [m]")
plt.legend()
plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as plt
from anaflow import theis


time = np.linspace(10, 200)
rad = [1, 5]

# Q(t) = Q * characteristic([0, T])
lap_kwargs = {"cond": 3, "cond_kw": {"a": 100}}
head = theis(
    time=time,
    rad=rad,
    storage=1e-4,
    transmissivity=1e-4,
    rate=-1e-4,
    lap_kwargs=lap_kwargs,
)

for i, step in enumerate(rad):
    plt.plot(time, head[:, i], label="Theis(r={})".format(step))

plt.title("The Stehfest algorithm is not suitable for this!")
plt.legend()
plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as plt
from anaflow import theis, ext_theis_2d


time_labels = ["10 s", "10 min", "10 h"]
time = [10, 600, 36000]       # 10s, 10min, 10h
rad = np.geomspace(0.05, 4)   # radius from the pumping well in [0, 4]
var = 0.5                     # variance of the log-transmissivity
len_scale = 10.0              # correlation length of the log-transmissivity
TG = 1e-4                     # the geometric mean of the transmissivity
TH = TG * np.exp(-var / 2.0)  # the harmonic mean of the transmissivity
S = 1e-4                      # storativity
rate = -1e-4                  # pumping rate

head_TG = theis(time, rad, S, TG, rate)
head_TH = theis(time, rad, S, TH, rate)
head_ef = ext_theis_2d(time, rad, S, TG, var, len_scale, rate)
time_ticks=[]
for i, step in enumerate(time):
    label_TG = "Theis($T_G$)" if i == 0 else None
    label_TH = "Theis($T_H$)" if i == 0 else None
    label_ef = "extended Theis" if i == 0 else None
    plt.plot(rad, head_TG[i], label=label_TG, color="C"+str(i), linestyle="--")
    plt.plot(rad, head_TH[i], label=label_TH, color="C"+str(i), linestyle=":")
    plt.plot(rad, head_ef[i], label=label_ef, color="C"+str(i))
    time_ticks.append(head_ef[i][-1])

plt.xlabel("r in [m]")
plt.ylabel("h in [m]")
plt.legend()
ylim = plt.gca().get_ylim()
plt.gca().set_xlim([0, rad[-1]])
ax2 = plt.gca().twinx()
ax2.set_yticks(time_ticks)
ax2.set_yticklabels(time_labels)
ax2.set_ylim(ylim)
plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as plt
from anaflow import theis, ext_theis_tpl


time_ticks = []
time_labels = ["10 s", "10 min", "10 h"]
time = [10, 600, 36000]      # 10s, 10min, 10h
rad = np.geomspace(0.05, 4)  # radial distance from the pumping well in [0, 4]
S = 1e-4                     # storage
KG = 1e-4                    # the geometric mean of the conductivity
len_scale = 20.0             # upper bound for the length scale
hurst = 0.5                  # hurst coefficient
var = 0.5                    # variance of the log-conductivity
rate = -1e-4                 # pumping rate
KH = KG * np.exp(-var / 2)   # the harmonic mean of the conductivity

head_KG = theis(time, rad, S, KG, rate)
head_KH = theis(time, rad, S, KH, rate)
head_ef = ext_theis_tpl(
    time=time,
    rad=rad,
    storage=S,
    cond_gmean=KG,
    len_scale=len_scale,
    hurst=hurst,
    var=var,
    rate=rate,
)
for i, step in enumerate(time):
    label_TG = "Theis($K_G$)" if i == 0 else None
    label_TH = "Theis($K_H$)" if i == 0 else None
    label_ef = "extended Theis TPL 2D" if i == 0 else None
    plt.plot(rad, head_KG[i], label=label_TG, color="C"+str(i), linestyle="--")
    plt.plot(rad, head_KH[i], label=label_TH, color="C"+str(i), linestyle=":")
    plt.plot(rad, head_ef[i], label=label_ef, color="C"+str(i))
    time_ticks.append(head_ef[i][-1])

plt.xscale("log")
plt.xlabel("r in [m]")
plt.ylabel("h in [m]")
plt.legend()
ylim = plt.gca().get_ylim()
plt.gca().set_xlim([rad[0], rad[-1]])
ax2 = plt.gca().twinx()
ax2.set_yticks(time_ticks)
ax2.set_yticklabels(time_labels)
ax2.set_ylim(ylim)
plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as plt
from anaflow import theis, thiem


time = [10, 100, 1000]
rad = np.geomspace(0.1, 10)
r_ref = 10.0

head_ref = theis(time, np.full_like(rad, r_ref), storage=1e-3, transmissivity=1e-4, rate=-1e-4)
head1 = theis(time, rad, storage=1e-3, transmissivity=1e-4, rate=-1e-4) - head_ref
head2 = theis(time, rad, storage=1e-3, transmissivity=1e-4, rate=-1e-4, r_bound=r_ref)
head3 = thiem(rad, r_ref, transmissivity=1e-4, rate=-1e-4)

for i, step in enumerate(time):
    label_1 = "Theis quasi steady" if i == 0 else None
    label_2 = "Theis bounded" if i == 0 else None
    plt.plot(rad, head1[i], label=label_1, color="C"+str(i), linestyle="--")
    plt.plot(rad, head2[i], label=label_2, color="C"+str(i))

plt.plot(rad, head3, label="Thiem", color="k", linestyle=":")

plt.xlabel("r in [m]")
plt.ylabel("h in [m]")
plt.legend()
plt.tight_layout()
plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# anaflow documentation build configuration file, created by
# sphinx-quickstart on Fri Jan  5 14:20:43 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# NOTE:
# pip install sphinx_rtd_theme
# is needed in order to build the documentation
import datetime
from anaflow import __version__ as ver


def skip(app, what, name, obj, skip, options):
    if name in ["__call__"]:
        return False
    return skip


def setup(app):
    app.connect("autodoc-skip-member", skip)


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.imgmath",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",  # parameters look better than with numpydoc only
    "numpydoc",
]

# autosummaries from source-files
autosummary_generate = True
# dont show __init__ docstring
autoclass_content = "class"
# sort class members
autodoc_member_order = "groupwise"
# autodoc_member_order = 'bysource'

# don't add full path to module
add_module_names = False

# Notes in boxes
napoleon_use_admonition_for_notes = True
# Attributes like parameters
# napoleon_use_ivar = True
# this is a nice class-doc layout
numpydoc_show_class_members = True
# class members have no separate file, so they are not in a toctree
numpydoc_class_members_toctree = False
# for the covmodels alot of classmembers show up...
numpydoc_show_inherited_class_members = True
# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
# --> this is the sitemap (or content-list in latex -> needs a heading)
# for html: the quickstart (in index.rst)
# gets the "index.html" and is therefore opened first
master_doc = "contents"

# General information about the project.
curr_year = datetime.datetime.now().year
project = "AnaFlow"
copyright = "2019 - {}, Sebastian Mueller".format(curr_year)
author = "Sebastian Mueller"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ver
# The full version, including alpha/beta/rc tags.
release = ver

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    #    'canonical_url': '',
    #    'analytics_id': '',
    "logo_only": False,
    "display_version": True,
    "prev_next_buttons_location": "top",
    #    'style_external_links': False,
    #    'vcs_pageview_mode': '',
    # Toc options
    "collapse_navigation": False,
    "sticky_navigation": True,
    "navigation_depth": 4,
    "includehidden": True,
    "titles_only": False,
}
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "**": [
        "relations.html",  # needs 'show_related': True theme option to display
        "searchbox.html",
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "AnaFlowdoc"
# logos for the page
html_logo = "pics/Anaflow_150.png"
html_favicon = "pics/Anaflow.ico"


# -- Options for LaTeX output ---------------------------------------------
# latex_engine = 'lualatex'
# logo to big
latex_logo = "pics/Anaflow_150.png"

# latex_show_urls = 'footnote'
# http://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-latex-output
latex_elements = {
    "preamble": r"""
\setcounter{secnumdepth}{1}
\setcounter{tocdepth}{2}
\pagestyle{fancy}
""",
    "pointsize": "10pt",
    "papersize": "a4paper",
    "fncychap": "\\usepackage[Glenn]{fncychap}",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "anaflow.tex",
        "AnaFlow Documentation",
        "Sebastian Mueller",
        "manual",
    )
]
# latex_use_parts = True

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "AnaFlow", "AnaFlow Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "AnaFlow",
        "AnaFlow Documentation",
        author,
        "AnaFlow",
        "Analytical solutions for the groundwater flow equation",
        "Miscellaneous",
    )
]

suppress_warnings = [
    "image.nonlocal_uri",
    # 'app.add_directive',  # this evtl. suppresses the numpydoc induced warning
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "Python 3.6": ("https://docs.python.org/3.6", None),
    "Python": ("https://docs.python.org/", None),
    "NumPy": ("http://docs.scipy.org/doc/numpy/", None),
    "SciPy": ("http://docs.scipy.org/doc/scipy/reference", None),
}



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as plt
from anaflow import ext_theis_tpl, ext_thiem_tpl

time = 1e4                   # time point for steady state
rad = np.geomspace(0.1, 10)  # radius from the pumping well in [0, 4]
r_ref = 10.0                 # reference radius
KG = 1e-4                    # the geometric mean of the transmissivity
len_scale = 5.0              # correlation length of the log-transmissivity
hurst = 0.5                  # hurst coefficient
var = 0.5                    # variance of the log-transmissivity
rate = -1e-4                 # pumping rate
dim = 1.5                    # using a fractional dimension

head1 = ext_thiem_tpl(rad, r_ref, KG, len_scale, hurst, var, dim=dim, rate=rate)
head2 = ext_theis_tpl(time, rad, 1e-4, KG, len_scale, hurst, var, dim=dim, rate=rate, r_bound=r_ref)

plt.plot(rad, head1, label="Ext Thiem TPL")
plt.plot(rad, head2, label="Ext Theis TPL (t={})".format(time), linestyle="--")

plt.xlabel("r in [m]")
plt.ylabel("h in [m]")
plt.legend()
plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
"""
This is the unittest of AnaFlow.
"""

import unittest
import numpy as np
import anaflow as af

# import matplotlib.pyplot as plt


def inc(arr, delta=0.0):
    return np.all(arr[:-1] <= arr[1:] + delta)


def dec(arr, delta=0.0):
    return np.all(arr[:-1] >= arr[1:] - delta)


def less(arr1, arr2, delta=0.0):
    return np.all(arr1 <= arr2 + delta)


class TestAnaFlow(unittest.TestCase):
    def setUp(self):
        self.delta = 1e-5
        self.t_gmean = 1e-4
        self.var = 0.5
        self.t_hmean = self.t_gmean * np.exp(-self.var / 2.0)
        self.anis = 0.4
        self.hurst = 0.7
        self.frac_dim = 2.3
        self.len_scale = 10.0
        self.lat_ext = 2.5
        self.r_ref = 20.0
        self.h_ref = -1.0
        self.rate = -1e-4
        self.storage = 1e-3
        self.rad = np.arange(1, 11, dtype=float)
        self.time = np.array([1e1, 1e2, 1e3, 1e4], dtype=float)

    def test_homogeneous(self):
        steady = af.thiem(
            rad=self.rad,
            r_ref=self.r_ref,
            transmissivity=self.t_gmean,
            rate=self.rate,
            h_ref=self.h_ref,
        )
        transient = af.theis(
            time=self.time,
            rad=self.rad,
            storage=self.storage,
            transmissivity=self.t_gmean,
            rate=self.rate,
            r_bound=self.r_ref,
            h_bound=self.h_ref,
        )
        self.assertTrue(inc(steady, self.delta))
        for rad_arr in transient:
            self.assertTrue(inc(rad_arr, self.delta))
        for time_arr in transient.T:
            self.assertTrue(dec(time_arr, self.delta))
        self.assertAlmostEqual(
            np.abs(np.max(steady - transient[-1, :])), 0.0, places=4
        )

    def test_ext_2d(self):
        steady = af.ext_thiem_2d(
            rad=self.rad,
            r_ref=self.r_ref,
            trans_gmean=self.t_gmean,
            var=self.var,
            len_scale=self.len_scale,
            rate=self.rate,
            h_ref=self.h_ref,
        )
        transient = af.ext_theis_2d(
            time=self.time,
            rad=self.rad,
            storage=self.storage,
            trans_gmean=self.t_gmean,
            var=self.var,
            len_scale=self.len_scale,
            rate=self.rate,
            r_bound=self.r_ref,
            h_bound=self.h_ref,
        )
        self.assertTrue(inc(steady, self.delta))
        for rad_arr in transient:
            self.assertTrue(inc(rad_arr, self.delta))
        for time_arr in transient.T:
            self.assertTrue(dec(time_arr, self.delta))
        self.assertAlmostEqual(
            np.abs(np.max(steady - transient[-1, :])), 0.0, places=4
        )

    def test_ext_3d(self):
        steady = af.ext_thiem_3d(
            rad=self.rad,
            r_ref=self.r_ref,
            cond_gmean=self.t_gmean,
            var=self.var,
            len_scale=self.len_scale,
            anis=self.anis,
            lat_ext=self.lat_ext,
            rate=self.rate,
            h_ref=self.h_ref,
        )
        transient = af.ext_theis_3d(
            time=self.time,
            rad=self.rad,
            storage=self.storage,
            cond_gmean=self.t_gmean,
            var=self.var,
            len_scale=self.len_scale,
            anis=self.anis,
            lat_ext=self.lat_ext,
            rate=self.rate,
            r_bound=self.r_ref,
            h_bound=self.h_ref,
        )
        self.assertTrue(inc(steady, self.delta))
        for rad_arr in transient:
            self.assertTrue(inc(rad_arr, self.delta))
        for time_arr in transient.T:
            self.assertTrue(dec(time_arr, self.delta))
        self.assertAlmostEqual(
            np.abs(np.max(steady - transient[-1, :])), 0.0, places=4
        )

    def test_neuman(self):
        steady = af.neuman2004_steady(
            rad=self.rad,
            r_ref=self.r_ref,
            trans_gmean=self.t_gmean,
            var=self.var,
            len_scale=self.len_scale,
            rate=self.rate,
            h_ref=self.h_ref,
        )
        transient = af.neuman2004(
            time=self.time,
            rad=self.rad,
            storage=self.storage,
            trans_gmean=self.t_gmean,
            var=self.var,
            len_scale=self.len_scale,
            rate=self.rate,
            r_bound=self.r_ref,
            h_bound=self.h_ref,
        )
        self.assertTrue(inc(steady, self.delta))
        for rad_arr in transient:
            self.assertTrue(inc(rad_arr, self.delta))
        for time_arr in transient.T:
            self.assertTrue(dec(time_arr, self.delta))
        self.assertAlmostEqual(
            np.abs(np.max(steady - transient[-1, :])), 0.0, places=4
        )

    def test_tpl(self):
        steady = af.ext_thiem_tpl(
            rad=self.rad,
            r_ref=self.r_ref,
            cond_gmean=self.t_gmean,
            var=self.var,
            len_scale=self.len_scale,
            lat_ext=self.lat_ext,
            hurst=self.hurst,
            dim=self.frac_dim,
            rate=self.rate,
            h_ref=self.h_ref,
        )
        transient = af.ext_theis_tpl(
            time=self.time,
            rad=self.rad,
            storage=self.storage,
            cond_gmean=self.t_gmean,
            var=self.var,
            len_scale=self.len_scale,
            lat_ext=self.lat_ext,
            hurst=self.hurst,
            dim=self.frac_dim,
            rate=self.rate,
            r_bound=self.r_ref,
            h_bound=self.h_ref,
        )
        self.assertTrue(inc(steady, self.delta))
        for rad_arr in transient:
            self.assertTrue(inc(rad_arr, self.delta))
        for time_arr in transient.T:
            self.assertTrue(dec(time_arr, self.delta))
        self.assertAlmostEqual(
            np.abs(np.max(steady - transient[-1, :])), 0.0, places=3
        )

    def test_tpl_3d(self):
        steady = af.ext_thiem_tpl_3d(
            rad=self.rad,
            r_ref=self.r_ref,
            cond_gmean=self.t_gmean,
            var=self.var,
            len_scale=self.len_scale,
            lat_ext=self.lat_ext,
            hurst=self.hurst,
            anis=self.anis,
            rate=self.rate,
            h_ref=self.h_ref,
        )
        transient = af.ext_theis_tpl_3d(
            time=self.time,
            rad=self.rad,
            storage=self.storage,
            cond_gmean=self.t_gmean,
            var=self.var,
            len_scale=self.len_scale,
            lat_ext=self.lat_ext,
            hurst=self.hurst,
            anis=self.anis,
            rate=self.rate,
            r_bound=self.r_ref,
            h_bound=self.h_ref,
        )
        self.assertTrue(inc(steady, self.delta))
        for rad_arr in transient:
            self.assertTrue(inc(rad_arr, self.delta))
        for time_arr in transient.T:
            self.assertTrue(dec(time_arr, self.delta))
        self.assertAlmostEqual(
            np.abs(np.max(steady - transient[-1, :])), 0.0, places=2
        )

        # plt.plot(self.rad, steady)
        # for rad_arr in transient:
        #     plt.plot(self.rad, rad_arr, dashes=[1, 1])
        # plt.show()


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""AnaFlow - analytical solutions for the groundwater-flow equation."""

import os
from setuptools import setup, find_packages


HERE = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(HERE, "README.md"), encoding="utf-8") as f:
    README = f.read()
with open(os.path.join(HERE, "requirements.txt"), encoding="utf-8") as f:
    REQ = f.read().splitlines()
with open(os.path.join(HERE, "requirements_setup.txt"), encoding="utf-8") as f:
    REQ_SETUP = f.read().splitlines()
with open(os.path.join(HERE, "requirements_test.txt"), encoding="utf-8") as f:
    REQ_TEST = f.read().splitlines()
with open(
    os.path.join(HERE, "docs", "requirements_doc.txt"), encoding="utf-8"
) as f:
    REQ_DOC = f.read().splitlines()

REQ_DEV = REQ_SETUP + REQ_TEST + REQ_DOC

DOCLINE = __doc__.split("\n")[0]
CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Intended Audience :: End Users/Desktop",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Operating System :: MacOS",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: Microsoft",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Topic :: Scientific/Engineering",
    "Topic :: Software Development",
    "Topic :: Utilities",
]

setup(
    name="anaflow",
    description=DOCLINE,
    long_description=README,
    long_description_content_type="text/markdown",
    maintainer="Sebastian Mueller",
    maintainer_email="sebastian.mueller@ufz.de",
    author="Sebastian Mueller",
    author_email="sebastian.mueller@ufz.de",
    url="https://github.com/GeoStat-Framework/AnaFlow",
    license="MIT",
    classifiers=CLASSIFIERS,
    platforms=["Windows", "Linux", "Mac OS-X"],
    include_package_data=True,
    python_requires=">=3.5",
    use_scm_version={
        "relative_to": __file__,
        "write_to": "anaflow/_version.py",
        "write_to_template": "__version__ = '{version}'",
        "local_scheme": "no-local-version",
        "fallback_version": "0.0.0.dev0",
    },
    install_requires=REQ,
    setup_requires=REQ_SETUP,
    extras_require={"doc": REQ_DOC, "test": REQ_TEST, "dev": REQ_DEV},
    packages=find_packages(exclude=["tests*", "docs*"]),
)



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as plt
from anaflow import ext_thiem_2d, ext_grf_steady
from anaflow.tools.coarse_graining import T_CG


rad = np.geomspace(0.05, 4)  # radius from the pumping well in [0, 4]
r_ref = 2.0                  # reference radius
var = 0.5                    # variance of the log-transmissivity
len_scale = 10.0             # correlation length of the log-transmissivity
TG = 1e-4                    # the geometric mean of the transmissivity
rate = -1e-4                 # pumping rate

head1 = ext_thiem_2d(rad, r_ref, TG, var, len_scale, rate)
head2 = ext_grf_steady(rad, r_ref, T_CG, rate=rate, trans_gmean=TG, var=var, len_scale=len_scale)

plt.plot(rad, head1, label="Ext Thiem 2D")
plt.plot(rad, head2, label="grf(T_CG)", linestyle="--")

plt.xlabel("r in [m]")
plt.ylabel("h in [m]")
plt.legend()
plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
import numpy as np
from scipy.special import erf
from matplotlib import pyplot as plt
import matplotlib.gridspec as gridspec
from anaflow import theis


time = np.geomspace(1, 600)
rad = [1, 5]

# Q(t) = Q * erf(t / a)
a = 120
lap_kwargs = {"cond": 4, "cond_kw": {"a": a}}
head1 = theis(
    time=time,
    rad=rad,
    storage=1e-4,
    transmissivity=1e-4,
    rate=-1e-4,
    lap_kwargs=lap_kwargs,
)
head2 = theis(
    time=time,
    rad=rad,
    storage=1e-4,
    transmissivity=1e-4,
    rate=-1e-4,
)
gs = gridspec.GridSpec(2, 1, height_ratios=[1, 3])
ax1 = plt.subplot(gs[0])
ax2 = plt.subplot(gs[1], sharex=ax1)

for i, step in enumerate(rad):
    ax2.plot(
        time,
        head1[:, i],
        color="C" + str(i),
        label="accruing Theis(r={})".format(step),
    )
    ax2.plot(
        time,
        head2[:, i],
        color="C" + str(i),
        label="constant Theis(r={})".format(step),
        linestyle="--"
    )
ax1.plot(time, 1e-4 * erf(time / a), label="accruing Q")
ax2.set_xlabel("t in [s]")
ax2.set_ylabel("h in [m]")
ax1.set_ylabel(r"|Q| in [$\frac{m^3}{s}$]")
ax1.legend()
ax2.legend()
plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
import numpy as np
from matplotlib import pyplot as plt
from anaflow import theis, ext_theis_3d
from anaflow.tools.special import aniso


time_labels = ["10 s", "10 min", "10 h"]
time = [10, 600, 36000]                  # 10s, 10min, 10h
rad = np.geomspace(0.05, 4)              # radial distance from the pumping well in [0, 4]
var = 0.5                                # variance of the log-conductivity
len_scale = 10.0                         # correlation length of the log-conductivity
anis = 0.75                              # anisotropy ratio of the log-conductivity
KG = 1e-4                                # the geometric mean of the conductivity
Kefu = KG*np.exp(var*(0.5-aniso(anis)))  # the effective conductivity for uniform flow
KH = KG*np.exp(-var/2.0)                 # the harmonic mean of the conductivity
S = 1e-4                                 # storage
rate = -1e-4                             # pumping rate
L = 1.0                                  # vertical extend of the aquifer

head_Kefu = theis(time, rad, S, Kefu*L, rate)
head_KH = theis(time, rad, S, KH*L, rate)
head_ef = ext_theis_3d(time, rad, S, KG, var, len_scale, anis, L, rate)
time_ticks=[]
for i, step in enumerate(time):
    label_TG = "Theis($K_{efu}$)" if i == 0 else None
    label_TH = "Theis($K_H$)" if i == 0 else None
    label_ef = "extended Theis 3D" if i == 0 else None
    plt.plot(rad, head_Kefu[i], label=label_TG, color="C"+str(i), linestyle="--")
    plt.plot(rad, head_KH[i], label=label_TH, color="C"+str(i), linestyle=":")
    plt.plot(rad, head_ef[i], label=label_ef, color="C"+str(i))
    time_ticks.append(head_ef[i][-1])

plt.xlabel("r in [m]")
plt.ylabel("h in [m]")
plt.legend()
ylim = plt.gca().get_ylim()
plt.gca().set_xlim([0, rad[-1]])
ax2 = plt.gca().twinx()
ax2.set_yticks(time_ticks)
ax2.set_yticklabels(time_labels)
ax2.set_ylim(ylim)
plt.tight_layout()
plt.show()



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing the extended GRF Model.

.. currentmodule:: anaflow.flow.ext_grf_model

The following functions are provided

.. autosummary::
   ext_grf
   ext_grf_steady
"""
# pylint: disable=C0103
import numpy as np
from scipy.integrate import quad as integ

from anaflow.tools.laplace import get_lap_inv
from anaflow.flow.laplace import grf_laplace
from anaflow.tools.special import Shaper, sph_surf

__all__ = ["ext_grf", "ext_grf_steady"]


def ext_grf(
    time,
    rad,
    S_part,
    K_part,
    R_part,
    dim=2,
    lat_ext=1.0,
    rate=-1e-4,
    h_bound=0.0,
    K_well=None,
    struc_grid=True,
    lap_kwargs=None,
):
    """
    The extended "General radial flow" model for transient flow.

    The general radial flow (GRF) model by Barker introduces an arbitrary
    dimension for radial groundwater flow. We introduced the possibility to
    define radial dependet conductivity and storage values.

    This solution is based on the grf model presented in [Barker88]_.

    Parameters
    ----------
    time : :class:`numpy.ndarray`
        Array with all time-points where the function should be evaluated
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    S_part : :class:`numpy.ndarray`
        Given storativity values for each disk
    K_part : :class:`numpy.ndarray`
        Given conductivity values for each disk
    R_part : :class:`numpy.ndarray`
        Given radii separating the disks (including r_well and r_bound).
    dim : :class:`float`, optional
        Fractional dimension of the aquifer. Default: ``2.0``
    lat_ext : :class:`float`, optional
        Lateral extend of the aquifer. Default: ``1.0``
    rate : :class:`float`, optional
        Pumpingrate at the well. Default: -1e-4
    h_bound : :class:`float`, optional
        Reference head at the outer boundary `R_part[-1]`. Default: ``0.0``
    K_well : :class:`float`, optional
        Conductivity at the well. Default: ``K_part[0]``
    struc_grid : :class:`bool`, optional
        If this is set to ``False``, the `rad` and `time` array will be merged
        and interpreted as single, r-t points. In this case they need to have
        the same shapes. Otherwise a structured r-t grid is created.
        Default: ``True``
    lap_kwargs : :class:`dict` or :any:`None` optional
        Dictionary for :any:`get_lap_inv` containing `method` and
        `method_dict`. The default is equivalent to
        ``lap_kwargs = {"method": "stehfest", "method_dict": None}``.
        Default: :any:`None`

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all heads at the given radii and time-points.

    References
    ----------
    .. [Barker88] Barker, J.
       ''A generalized radial flow model for hydraulic tests
       in fractured rock.'',
       Water Resources Research 24.10, 1796-1804, 1988
    """
    Input = Shaper(time, rad, struc_grid)
    lap_kwargs = {} if lap_kwargs is None else lap_kwargs
    # write the paramters in kwargs to use the stehfest-algorithm
    kwargs = {
        "rad": rad,
        "R_part": R_part,
        "S_part": S_part,
        "K_part": K_part,
        "dim": dim,
        "lat_ext": lat_ext,
        "rate": rate,
        "K_well": K_well,
    }
    kwargs.update(lap_kwargs)

    res = np.zeros((Input.time_no, Input.rad_no))
    # call the grf-model
    lap_inv = get_lap_inv(grf_laplace, **kwargs)
    res[Input.time_gz, :] = lap_inv(Input.time[Input.time_gz])
    res = Input.reshape(res)
    # add the reference head
    res += h_bound
    return res


def ext_grf_steady(
    rad,
    r_ref,
    conductivity,
    dim=2,
    lat_ext=1.0,
    rate=-1e-4,
    h_ref=0.0,
    arg_dict=None,
    **kwargs
):
    """
    The extended "General radial flow" model for steady flow.

    The general radial flow (GRF) model by Barker introduces an arbitrary
    dimension for radial groundwater flow. We introduced the possibility to
    define radial dependet conductivity.

    This solution is based on the grf model presented in [Barker88]_.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    r_ref : :class:`float`
        Radius of the reference head.
    conductivity : :class:`float` or :any:`callable`
        Conductivity. Either callable function taking kwargs or float.
    dim : :class:`float`, optional
        Fractional dimension of the aquifer. Default: ``2.0``
    lat_ext : :class:`float`, optional
        Lateral extend of the aquifer. Default: ``1.0``
    rate : :class:`float`, optional
        Pumpingrate at the well. Default: -1e-4
    h_ref : :class:`float`, optional
        Reference head at the reference-radius `r_ref`. Default: ``0.0``
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        conductivity function. Will be merged with ``**kwargs``.
        This is designed for overlapping keywords in ``grf_steady`` and
        ``conductivity``.
        Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the conductivity function.
        Will be merged with ``arg_dict``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all heads at the given radii and time-points.

    References
    ----------
    .. [Barker88] Barker, J.
       ''A generalized radial flow model for hydraulic tests
       in fractured rock.'',
       Water Resources Research 24.10, 1796-1804, 1988
    """
    arg_dict = {} if arg_dict is None else arg_dict
    kwargs.update(arg_dict)
    Input = Shaper(rad=rad)
    q_fac = rate / (sph_surf(dim) * lat_ext ** (3.0 - dim))  # pumping factor
    if not r_ref > 0.0:
        raise ValueError("The reference radius needs to be positive.")
    if not Input.rad_min > 0.0:
        raise ValueError("The given radii need to be positive.")
    if not dim > 0.0 or dim > 3.0:
        raise ValueError("The dimension needs to be positiv and <= 3.")
    if not lat_ext > 0.0:
        raise ValueError("The lateral extend needs to be positiv.")

    if callable(conductivity):
        res = np.zeros(Input.rad_no)

        def integrand(val):
            """Integrand."""
            return val ** (1 - dim) / conductivity(val, **kwargs)

        for ri, re in enumerate(Input.rad):
            res[ri] = integ(integrand, re, r_ref)[0]
    else:
        con = float(conductivity)
        if not con > 0:
            raise ValueError("The Conductivity needs to be positive.")
        if np.isclose(dim, 2):
            res = np.log(r_ref / Input.rad) / con
        else:
            res = (
                (r_ref ** (2 - dim) - Input.rad ** (2 - dim)) / (2 - dim) / con
            )

    res = Input.reshape(res)
    # rescale by pumping rate
    res *= q_fac
    # add the reference head
    res += h_ref
    return res


if __name__ == "__main__":
    import doctest

    doctest.testmod()
